# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""GP-Bandit using a Flax model and a TFP Gaussian Process.

A Python implementation of Google Vizier's GP-Bandit algorithm.
"""
# pylint: disable=logging-fstring-interpolation, g-long-lambda

import copy
import datetime
import functools
import json
import random
from typing import Any, Optional, Sequence, Union

from absl import logging
import attr
import jax
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.algorithms.designers.gp import output_warpers
from vizier._src.algorithms.optimizers import eagle_strategy as es
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import gp_bandit_utils
from vizier._src.jax import predictive_fns
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import tuned_gp_models
from vizier._src.jax.optimizers import optimizers
from vizier.pyvizier import converters
from vizier.pyvizier.converters import feature_mapper
from vizier.pyvizier.converters import padding
from vizier.utils import json_utils
from vizier.utils import profiler


# Define top-level JIT-ed versions of some library functions. See
# `gp_bandit_utils.py` for function documentation.
jit_precompute_cholesky = profiler.record_runtime(
    jax.jit(gp_bandit_utils.precompute_cholesky, static_argnames=('model',)),
    name_prefix='VizierGPBandit',
    name='precompute_cholesky',
)

jit_vmap_precompute_cholesky = profiler.record_runtime(
    jax.jit(
        jax.vmap(gp_bandit_utils.precompute_cholesky, in_axes=(None, None, 0)),
        static_argnames=('model',),
    ),
    name_prefix='VizierGPBandit',
    name='vmap_precompute_cholesky',
)

jit_optimize_acquisition = profiler.record_runtime(
    jax.jit(
        gp_bandit_utils.optimize_acquisition,
        static_argnames=('count', 'model', 'use_vmap', 'mapper'),
    ),
    name_prefix='VizierGPBandit',
    name='optimize_acquisition',
)


@attr.define(auto_attribs=False)
class VizierGPBandit(vza.Designer, vza.Predictor):
  """GP-Bandit using a Flax model.

  A minimal example of creating this designer:
  problem = vz.ProblemStatement(...)  # Configure a minimal problem statement.
  designer = VizierGPBandit(problem)

  Optionally set other attributes to change the defaults, e.g.:
  problem = vz.ProblemStatement(...)  # Configure a minimal problem statement.
  designer = VizierGPBandit(problem, use_trust_region=False)

  Attributes:
    problem: Must be a flat study with a single metric.
    acquisition_optimizer: Typically either a designer wrapped as an optimizer
      or a batched optimizer (like Eagle).
    ard_optimizer: An optimizer which should return a batch of hyperparameters
      to be ensembled.
    num_seed_trials: If greater than zero, first trial is the center of the
      search space. Afterwards, uses quasirandom until this number of trials are
      observed.
    acquisition_builder: An acquisition builder instance specifying the
      acqusition function to use.
    use_trust_region: Uses trust region to constrain initial exploration.
    rng: If not set, uses random numbers.
    metadata_ns: Metadata namespace that this designer writes to.
  """

  _problem: vz.ProblemStatement = attr.field(kw_only=False)
  _acquisition_optimizer_factory: vb.VectorizedOptimizerFactory = attr.field(
      kw_only=True,
      factory=lambda: VizierGPBandit.default_acquisition_optimizer_factory,
  )
  _ard_optimizer: optimizers.Optimizer[types.ParameterDict] = attr.field(
      factory=lambda: VizierGPBandit.default_ard_optimizer_noensemble,
      kw_only=True,
  )
  _num_seed_trials: int = attr.field(default=1, kw_only=True)
  _acquisition_function: acquisitions.AcquisitionFunction = attr.field(
      factory=acquisitions.UCB, kw_only=True
  )
  _use_categorical_kernel: bool = attr.field(default=False, kw_only=True)
  # Whether to pad all inputs, and what type of schedule to use. This is to
  # ensure fewer JIT compilation passes.
  _padding_schedule: Optional[padding.PaddingSchedule] = attr.field(
      default=None, kw_only=True
  )
  _use_trust_region: bool = attr.field(default=True, kw_only=True)
  _rng: jax.random.KeyArray = attr.field(
      factory=lambda: jax.random.PRNGKey(random.getrandbits(32)), kw_only=True
  )
  _metadata_ns: str = attr.field(
      default='oss_gp_bandit', kw_only=True, init=False
  )

  # ------------------------------------------------------------------
  # Internal attributes which should not be set by callers.
  # ------------------------------------------------------------------
  _trials: list[vz.Trial] = attr.field(factory=list, init=False)
  # The number of trials that have been incorporated
  # into the designer state (Cholesky decomposition, ARD).
  _incorporated_trials_count: int = attr.field(
      default=0, kw_only=True, init=False
  )
  _acquisition_optimizer: vb.VectorizedOptimizer = attr.field(init=False)
  _output_warper_pipeline: output_warpers.OutputWarperPipeline = attr.field(
      init=False
  )
  _feature_mapper: Optional[
      feature_mapper.ContinuousCategoricalFeatureMapper
  ] = attr.field(init=False, default=None)
  _last_computed_state: types.GPState = attr.field(init=False)

  # ------------------------------------------------------------------
  # Below are class contants which are not attr fields.
  # ------------------------------------------------------------------
  # Only one of these optimizers will be used.
  # `default_ard_optimizer_ensemble` returns the best 5 parameter values for
  # ensembling, while `default_ard_optimizer_noensemble` returns only
  # the single best parameter value.
  default_ard_optimizer_ensemble = optimizers.JaxoptLbfgsB(
      random_restarts=8, best_n=5
  )
  default_ard_optimizer_noensemble = optimizers.JaxoptLbfgsB(
      random_restarts=4, best_n=1
  )
  default_acquisition_optimizer_factory = vb.VectorizedOptimizerFactory(
      strategy_factory=es.VectorizedEagleStrategyFactory()
  )

  def __attrs_post_init__(self):
    # Extra validations
    if self._problem.search_space.is_conditional:
      raise ValueError(f'{type(self)} does not support conditional search.')
    elif len(self._problem.metric_information) != 1:
      raise ValueError(f'{type(self)} works with exactly one metric.')
    if self._use_categorical_kernel and self._padding_schedule:
      raise ValueError(
          f'{type(self)} does not support padding with categorical kernel.'
          f' use_categorical_kernel = {self._use_categorical_kernel},'
          f' padding_schedule = {self._padding_schedule}.'
      )
    # Extra initializations.
    # Discrete parameters are continuified to account for their actual values.
    if self._padding_schedule:
      self._converter = (
          converters.PaddedTrialToArrayConverter.from_study_config(
              self._problem,
              scale=True,
              pad_oovs=True,
              padding_schedule=self._padding_schedule,
              max_discrete_indices=0,
              flip_sign_for_minimization_metrics=True,
          )
      )
    else:
      self._converter = converters.TrialToArrayConverter.from_study_config(
          self._problem,
          scale=True,
          pad_oovs=True,
          max_discrete_indices=0,
          flip_sign_for_minimization_metrics=True,
      )
    self._quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        self._problem.search_space
    )
    self._output_warper_pipeline = output_warpers.create_default_warper()
    if self._use_categorical_kernel:
      self._feature_mapper = feature_mapper.ContinuousCategoricalFeatureMapper(
          self._converter
      )
    self._acquisition_optimizer = self._acquisition_optimizer_factory(
        self._converter
    )
    acquisition_problem = copy.deepcopy(self._problem)
    if isinstance(
        self._acquisition_function, acquisitions.MultiAcquisitionFunction
    ):
      acquisition_config = vz.MetricsConfig()
      for k in self._acquisition_function.acquisition_fns.keys():
        acquisition_config.append(
            vz.MetricInformation(
                name=k,
                goal=vz.ObjectiveMetricGoal.MAXIMIZE,
            )
        )
    else:
      acquisition_config = vz.MetricsConfig(
          metrics=[
              vz.MetricInformation(
                  name='acquisition', goal=vz.ObjectiveMetricGoal.MAXIMIZE
              )
          ]
      )

    acquisition_problem.metric_information = acquisition_config
    self._acquisition_problem = acquisition_problem

  def _build_model(self, features):
    if self._use_categorical_kernel:
      return tuned_gp_models.VizierGaussianProcessWithCategorical.build_model(
          features
      )
    return tuned_gp_models.VizierGaussianProcess.build_model(features)

  @property
  def _use_vmap(self):
    """Returns whether ensemble of parameters is used which requires vmap."""
    # Derived classes of `optimizers.Optimizer` have a `best_n` property that
    # indicates whether the optimizer trains an ensemble of parameters and
    # therefore whether vmap should be used.
    # If the optimizer has a `best_n` property with value greater than 1, vmap
    # is used.
    return getattr(self._ard_optimizer, 'best_n', -1) > 1

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the list of completed trials."""
    del all_active
    self._trials.extend(copy.deepcopy(completed.trials))

  @property
  def _metric_info(self) -> vz.MetricInformation:
    return self._problem.metric_information.item()

  # TODO: Check the latency of `_generate_seed_trials` and look
  # into reducing it.
  @profiler.record_runtime(
      name_prefix='VizierGPBandit', name='generate_seed_trials'
  )
  def _generate_seed_trials(self, count: int) -> Sequence[vz.TrialSuggestion]:
    """Generate seed trials.

    The first seed trial is chosen as the search space center, the rest of the
    seed trials are chosen quasi-randomly.

    Arguments:
      count: The number of seed trials.

    Returns:
      The seed trials.
    """
    seed_suggestions = []
    if not self._trials:
      # TODO: Should track number of pending suggestions
      # so we don't suggest the center more than once.
      features = self._converter.to_features([])  # to extract shape.
      # NOTE: The code below assumes that a scaled value of 0.5 corresponds
      #   to the center of the feasible range.  This is true, but only by
      #   accident; ideally, we should get the center from the converters.
      parameters = self._converter.to_parameters(
          0.5 * np.ones([1, features.shape[1]])
      )[0]
      seed_suggestions.append(
          vz.TrialSuggestion(
              parameters, metadata=vz.Metadata({'seeded': 'center'})
          )
      )
    if (remaining_counts := count - len(seed_suggestions)) > 0:
      seed_suggestions.extend(
          self._quasi_random_sampler.suggest(remaining_counts)
      )
    return seed_suggestions

  @profiler.record_runtime(
      name_prefix='VizierGPBandit', name='convert_trials_to_arrays'
  )
  def _convert_trials_to_arrays(
      self, trials: Sequence[vz.Trial]
  ) -> tuple[
      Union[types.MaybePaddedArray, types.ContinuousAndCategoricalArray],
      types.MaybePaddedArray,
  ]:
    """Convert trials to scaled features and warped labels."""
    features, pre_labels = self._converter.to_xy(self._trials)
    logging.info(
        'Transforming the labels of shape %s. Features has shape: %s',
        pre_labels.shape,
        features.shape,
    )
    # Map to continuous and categorical.
    if self._use_categorical_kernel:
      features = self._feature_mapper.map(features)

    labels = pre_labels
    if self._padding_schedule:
      # Labels coming from the converter will have shape [P, 1], where P >= T
      # the number of trials.
      padded_labels = pre_labels.padded_array
      # Unpad labels to have shape [T, 1].
      # This is because the warper may take into account all labels, which may
      # be NaN.
      labels = padded_labels[: len(self._trials)]

    # Warp the output.
    labels = self._output_warper_pipeline.warp(labels)
    labels = labels.reshape([-1])

    # Pad back
    if self._padding_schedule:
      # Pad back to shape [P, 1].
      padded_labels = np.concatenate(
          [labels, padded_labels[len(self._trials) :, 0]], axis=0
      )
      labels = padding.PaddedArray(
          padded_array=padded_labels, is_missing=pre_labels.is_missing
      )
    logging.info('Transformed the labels. Now has shape: %s', labels.shape)
    return features, labels

  @profiler.record_runtime(
      name_prefix='VizierGPBandit', name='ard', also_log=True
  )
  def _find_best_model_params(
      self,
      model: sp.StochasticProcessModel,
      loss_fn: optimizers.LossFunction,
      data: types.StochasticProcessModelData,
      seed: jax.random.KeyArray,
  ) -> tuple[types.ParameterDict, Any]:
    """Perform ARD on the current model to find best model parameters."""
    # Run ARD.
    setup = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_setup,
            static_argnames=('model',),
        ),
        model=model,
        data=data,
    )
    constraints = sp.get_constraints(model)

    logging.info('Optimizing the loss function...')

    # The ARD optimizers JIT the train step/loop internally.
    return self._ard_optimizer(setup, loss_fn, seed, constraints=constraints)

  @profiler.record_runtime(name_prefix='VizierGPBandit', name='compute_state')
  def _compute_state(
      self,
  ) -> tuple[types.GPState, Optional[acquisitions.TrustRegion]]:
    """Compute the designer's state.

    Returns:
      GPBanditState object containing the designer's state.

    1. Convert trials to features and labels.
    2. Perform ARD to find best model parameters.
    3. Pre-compute the Cholesky decomposition.

    If no new trials were added since last call, no update will occur.
    """
    if len(self._trials) == self._incorporated_trials_count:
      # If there's no change in the number of completed trials, don't update
      # state. The assumption is that trials can't be removed.
      return self._last_computed_state

    self._incorporated_trials_count = len(self._trials)
    # Convert trials to Numpy arrays. Labels are warped.
    features, labels = self._convert_trials_to_arrays(self._trials)

    dimension_is_missing = None
    label_is_missing = None
    if self._padding_schedule:
      dimension_is_missing = features.is_missing[1]
      label_is_missing = labels.is_missing[0]
      features = features.padded_array
      labels = labels.padded_array
    # Update the model.
    # TODO: Add support for PaddedArrays instead of passing in
    # masks.
    data = types.StochasticProcessModelData(
        features=features,
        labels=labels,
        label_is_missing=label_is_missing,
        dimension_is_missing=dimension_is_missing,
    )
    model = self._build_model(features)
    # TODO: Avoid retracing vmapped loss when loss function API is
    # redesigned.
    loss_fn = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model,
        data=data,
        # For SGD, normalize the loss so we can use the same learning rate
        # regardless of the number of examples (see
        # `OptaxTrainWithRandomRestarts` docstring).
        normalize=isinstance(
            self._ard_optimizer, optimizers.OptaxTrainWithRandomRestarts
        ),
    )
    self._rng, ard_rng = jax.random.split(self._rng, 2)
    best_model_params, metrics = self._find_best_model_params(
        model, loss_fn, data, ard_rng
    )
    # Logging for debugging purposes.
    logging.info('Best model parameters: %s', best_model_params)

    # ARD optimizer metrics dicts are assumed to have a 'loss' field.
    logging.info('All losses: %s', metrics['loss'])
    logging.info('ARD best loss: %s', np.min(metrics['loss']))

    if self._use_vmap:
      precompute_cholesky_fn = jit_vmap_precompute_cholesky
    else:
      precompute_cholesky_fn = jit_precompute_cholesky
    # `pp_state` contains intermediates that are expensive to compute, depend
    # only on observed (not predictive) index points, and are needed to compute
    # the posterior predictive GP (i.e. the Cholesky factor of the kernel matrix
    # over observed index points). `pp_state` is passed to
    # `model.posterior_predictive` to avoid re-computing the intermediates
    # unnecessarily.
    _, pp_state = precompute_cholesky_fn(model, data, best_model_params)
    model_state = {'params': best_model_params, **pp_state}

    trust_region = None
    if self._use_trust_region:
      trust_region = acquisitions.TrustRegion.build(
          self._converter.output_specs,
          data=data,
      )
    state = types.GPState(data=data, model_state=model_state), trust_region
    self._last_computed_state = state
    return state

  def _optimize_acquisition(
      self,
      count: int,
      state: types.GPState,
      trust_region: Optional[acquisitions.TrustRegion] = None,
  ) -> list[vz.Trial]:
    start_time = datetime.datetime.now()
    acq_rng, self._rng = jax.random.split(self._rng)

    prior_features = vb.trials_to_sorted_array(self._trials, self._converter)
    model = self._build_model(state.data.features)
    suggestions = jit_optimize_acquisition(
        count=count,
        model=model,
        acquisition_fn=self._acquisition_function,
        optimizer=self._acquisition_optimizer,
        prior_features=prior_features,
        state=state,
        seed=acq_rng,
        use_vmap=self._use_vmap,
        trust_region=trust_region,
        mapper=self._feature_mapper,
    )

    logging.info(
        (
            'Optimization completed. Duration: %s. Evaluations: %s. Best'
            ' Results: %s'
        ),
        datetime.datetime.now() - start_time,
        (
            (
                self._acquisition_optimizer.max_evaluations
                // self._acquisition_optimizer.suggestion_batch_size
            )
            * self._acquisition_optimizer.suggestion_batch_size
        ),
        suggestions,
    )
    return vb.best_candidates_to_trials(suggestions, self._converter)

  @profiler.record_runtime(name_prefix='VizierGPBandit')
  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    logging.info('Suggest called with count=%d', count)
    if count > 1:
      logging.warning(
          'GAUSSIAN_PROCESS_BANDIT currently is not optimized for batched'
          ' suggestions. Suggestions in the batch are likely to be very'
          ' similar.'
      )
    suggest_start_time = datetime.datetime.now()
    if len(self._trials) < self._num_seed_trials:
      return self._generate_seed_trials(count)

    logging.info('Updating the designer state based on trials...')
    state, _ = self._compute_state()

    logging.info('Optimizing acquisition...')
    best_candidates = self._optimize_acquisition(count, state)

    # Convert best_candidates (in scaled space) into suggestions (in unscaled
    # space); also append debug information like model predictions.
    logging.info('Converting the optimization result into suggestions...')
    optimal_features = self._converter.to_features(best_candidates)  # [N, D]
    if self._padding_schedule:
      optimal_features = optimal_features.padded_array
    if self._use_categorical_kernel:
      optimal_features = self._feature_mapper.map(optimal_features)
    # Make predictions (in the warped space). [N]
    model = self._build_model(state.data.features)
    with profiler.record_runtime_context(
        'VizierGPBandit.predict_on_suggestions'
    ):
      predictions = jax.jit(
          predictive_fns.predict_on_array, static_argnames=('model', 'use_vmap')
      )(
          optimal_features,
          model=model,
          state=state,
          use_vmap=self._use_vmap,
      )
    predict_mean = predictions['mean']  # [N,]
    predict_stddev = predictions['stddev']  # [N,]
    # Possibly unpad predictions
    predict_mean = predict_mean[: len(self._trials)]
    predict_stddev = predict_stddev[: len(self._trials)]
    # Create suggestions, injecting the predictions as metadata for
    # debugging needs.
    suggestions = []
    for i, candidate in enumerate(best_candidates):
      metadata = candidate.metadata.ns(self._metadata_ns).ns('devinfo')
      acquisition = candidate.final_measurement.metrics.get_value(
          'acquisition', 'nan'
      )
      metadata['prediction_in_warped_y_space'] = json.dumps(
          {
              'mean': predict_mean[i],
              'stddev': predict_stddev[i],
              'acquisition': acquisition,
          },
          cls=json_utils.NumpyEncoder,
      )
      metadata['time_spent'] = f'{datetime.datetime.now() - suggest_start_time}'
      suggestions.append(
          vz.TrialSuggestion(candidate.parameters, metadata=candidate.metadata)
      )

    return suggestions

  @profiler.record_runtime(name_prefix='VizierGPBandit')
  def predict(
      self,
      trials: Sequence[vz.TrialSuggestion],
      rng: Optional[jax.random.KeyArray] = None,
      num_samples: Optional[int] = None,
  ) -> vza.Prediction:
    """Returns the mean and stddev for any given trials.

    The method performs sampling of the warped GP model, unwarp the samples and
    compute the empirical mean and stadard deviation as an apprixmation.

    Arguments:
      trials: The trials where the predictions will be made.
      rng: The sampling random key used for approximation.
      num_samples: The number of samples used for the approximation.

    Returns:
      The predictions in the specified trials.
    """
    if rng is None:
      rng = jax.random.PRNGKey(0)
    if num_samples is None:
      num_samples = 1000

    state, _ = self._compute_state()
    model = self._build_model(state.data.features)
    xs = self._converter.to_features(trials)
    if self._padding_schedule:
      xs = xs.padded_array[: len(trials), ...]
    if self._use_categorical_kernel:
      xs = self._feature_mapper.map(xs)

    samples = jax.jit(
        predictive_fns.sample_on_array,
        static_argnames=('model', 'use_vmap', 'num_samples'),
    )(
        xs,
        num_samples=num_samples,
        key=rng,
        model=model,
        state=state,
        use_vmap=self._use_vmap,
    )  # (num_samples, batch_size)
    unwarped_samples = None
    # TODO: vectorize output warping.
    for i in range(samples.shape[0]):
      unwarp_samples_ = self._output_warper_pipeline.unwarp(
          samples[i][..., np.newaxis]
      ).reshape(-1)
      if unwarped_samples is not None:
        unwarped_samples = np.vstack([unwarp_samples_, unwarped_samples])
      else:
        unwarped_samples = unwarp_samples_

    mean = np.mean(unwarped_samples, axis=0)
    stddev = np.std(unwarped_samples, axis=0)
    return vza.Prediction(mean=mean, stddev=stddev)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Base class for vectorized acquisition optimizers."""

import abc
from typing import Generic, Optional, Protocol, TypeVar, Union

import attr
from flax import struct
import jax
from jax import numpy as jnp
import numpy as np
from vizier import pyvizier as vz
from vizier._src.jax import types
from vizier.pyvizier import converters

_S = TypeVar('_S')  # A container of optimizer state that works as a Pytree.

ArrayConverter = Union[
    converters.TrialToArrayConverter, converters.PaddedTrialToArrayConverter
]


@struct.dataclass
class VectorizedStrategyResults:
  """Container for a vectorized strategy result."""

  features: types.Array
  rewards: types.Array


class VectorizedStrategy(abc.ABC, Generic[_S]):
  """Interface class to implement a pure vectorized strategy.

  The strategy is responsible for generating suggestions that will maximize the
  reward. The order of calls is important. It's expected to be used in
  'suggest','update', 'suggest', 'update', etc.
  """

  @abc.abstractmethod
  def init_state(
      self,
      seed: jax.random.KeyArray,
      prior_features: Optional[types.Array] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> _S:
    """Initialize the state.

    Arguments:
      seed: Random seed for state initialization.
      prior_features: (n_prior_features, features_count)
      prior_rewards: (n_prior_features, )

    Returns:
      initial_state:
    """

  @abc.abstractmethod
  def suggest(self, state: _S, seed: jax.random.KeyArray) -> jax.Array:
    """Generate new suggestions.

    Arguments:
      state: Optimizer state.
      seed: Random seed.

    Returns:
      suggested features: (batch_size, features_count)
    """

  @property
  @abc.abstractmethod
  def suggestion_batch_size(self) -> int:
    """The number of suggestions returned at every suggest call."""

  @abc.abstractmethod
  def update(
      self,
      state: _S,
      batch_features: types.Array,
      batch_rewards: types.Array,
      seed: jax.random.KeyArray,
  ) -> _S:
    """Update the strategy state with the results of the last suggestions.

    Arguments:
      state: Optimizer state.
      batch_features: (batch_size, features_count)
      batch_rewards: (batch_size, )
      seed: Random seed.
    """


class VectorizedStrategyFactory(Protocol):
  """Factory class to generate vectorized strategy.

  It's used in VectorizedOptimizer to create a new strategy every 'optimize'
  call.
  """

  def __call__(
      self,
      converter: ArrayConverter,
      *,
      suggestion_batch_size: int,
  ) -> VectorizedStrategy:
    """Create a new vectorized strategy.

    Arguments:
      converter: The trial to array converter.
      suggestion_batch_size: The number of trials to be evaluated at once.
    """
    ...


class ArrayScoreFunction(Protocol):
  """Protocol for scoring array of trials.

  This protocol is suitable for optimizing batch of candidates (each one with
  its own separate score).
  """

  def __call__(self, batched_array_trials: types.Array) -> types.Array:
    """Evaluates the array of batched trials.

    Arguments:
      batched_array_trials: Array of shape (batch_size, n_feature_dimensions).

    Returns:
      Array of shape (batch_size,).
    """


class ParallelArrayScoreFunction(Protocol):
  """Protocol for scoring array of parallel trials.

  This protocol is suitable for optimizing in parallel multiple candidates
  (e.g. qUCB).
  """

  def __call__(self, parallel_array_trials: types.Array) -> types.Array:
    """Evaluates the array of batched trials.

    Arguments:
      parallel_array_trials: Array of shape (batch_size, n_parallel_candidates,
        n_feature_dimensions).

    Returns:
      Array of shape (batch_size).
    """


@struct.dataclass
class VectorizedOptimizer:
  """Vectorized strategy optimizer.

  The optimizer is stateless and will create a new vectorized strategy at the
  beginning of every 'optimize' call.

  The optimizer is responsible for running the iterative optimization process
  using the vectorized strategy, which consists of:
  1. Ask the strategy for suggestions.
  2. Evaluate the suggestions to get rewards.
  3. Tell the strategy about the rewards of its suggestion, so the strategy can
  update its internal state.

  The optimization process will terminate when the time limit or the total
  objective function evaluations limit has exceeded.

  Attributes:
    strategy: A factory for generating new strategy.
    n_feature_dimensions_with_padding: Number of feature dimensions including
      padding.
    n_feature_dimensions: Number of feature dimensions (less than or equal to
      `n_feature_dimensions_with_padding`).
    suggestion_batch_size: Number of suggested points returned at each call.
    max_evaluations: The maximum number of objective function evaluations.
  """

  strategy: VectorizedStrategy
  n_feature_dimensions: int = struct.field(pytree_node=False)
  n_feature_dimensions_with_padding: int = struct.field(pytree_node=False)
  suggestion_batch_size: int = struct.field(pytree_node=False, default=25)
  max_evaluations: int = struct.field(pytree_node=False, default=75_000)

  def __call__(
      self,
      score_fn: ArrayScoreFunction,
      *,
      count: int = 1,
      prior_features: Optional[types.Array] = None,
      seed: Optional[int] = None,
  ) -> VectorizedStrategyResults:
    """Optimize the objective function.

    The ask-evaluate-tell optimization procedure that runs until the allocated
    time or evaluations count exceeds.

    The number of suggestions is determined by the strategy, which is the
    `suggestion_count` property.

    The converter should be the same one used to convert trials to arrays in the
    format that the objective function expects to receive. It's used to convert
    backward the strategy's best array candidates to trial candidates.
    In addition, the converter is passed to the strategy so it could be aware
    of the type associated with each array index, and which indices are part of
    the same CATEGORICAL parameter one-hot encoding.

    The optimization stops when either of 'max_evaluations' or 'max_duration' is
    reached.

    Arguments:
      score_fn: A callback that expects 2D Array with dimensions (batch_size,
        features_count) and returns a 1D Array (batch_size,).
      count: The number of suggestions to generate.
      prior_features: Completed trials to be used for knowledge transfer. When
        the optimizer is used to optimize a designer's acquisition function, the
        prior trials are the previous designer suggestions provided in the
        ordered they were suggested.
      seed: The seed to use in the random generator.

    Returns:
      The best trials found in the optimization.
    """
    seed = jax.random.PRNGKey(0) if seed is None else seed

    input_is_padded = (
        self.n_feature_dimensions_with_padding > self.n_feature_dimensions
    )
    dimension_is_missing = None
    if input_is_padded:
      dimension_is_missing = np.array(
          [False] * self.n_feature_dimensions
          + [True]
          * (self.n_feature_dimensions_with_padding - self.n_feature_dimensions)
      )
    prior_rewards = None
    if prior_features is not None:
      prior_rewards = score_fn(prior_features).reshape(-1)

    def _optimization_one_step(_, args):
      state, best_results, seed = args
      suggest_seed, update_seed, new_seed = jax.random.split(seed, num=3)
      new_features = self.strategy.suggest(state, suggest_seed)
      # Ensure masking out padded dimensions in new features.
      if input_is_padded:
        new_features = jnp.where(
            dimension_is_missing, jnp.zeros_like(new_features), new_features
        )
      # We assume `score_fn` is aware of padded dimensions.
      new_rewards = score_fn(new_features)
      new_state = self.strategy.update(
          state, new_features, new_rewards, update_seed
      )
      new_best_results = self._update_best_results(
          best_results, count, new_features, new_rewards
      )
      return new_state, new_best_results, new_seed

    init_seed, loop_seed = jax.random.split(seed)
    init_best_results = VectorizedStrategyResults(
        rewards=-jnp.inf * jnp.ones([count]),
        features=jnp.zeros([count, self.n_feature_dimensions_with_padding]),
    )
    init_args = (
        self.strategy.init_state(
            init_seed,
            prior_features=prior_features,
            prior_rewards=prior_rewards,
        ),
        init_best_results,
        loop_seed,
    )
    _, best_results, _ = jax.lax.fori_loop(
        0,
        self.max_evaluations // self.suggestion_batch_size,
        _optimization_one_step,
        init_args,
    )

    return best_results

  def _update_best_results(
      self,
      best_results: VectorizedStrategyResults,
      count: int,
      batch_features: jax.Array,
      batch_rewards: jax.Array,
  ) -> VectorizedStrategyResults:
    """Update the best results the optimizer seen thus far.

    The best results are kept in a heap to efficiently maintain the top 'count'
    results throughout the optimizer run.

    Arguments:
      best_results: A heap storing the best results seen thus far. Implemented
        as a list of maximum size of 'count'.
      count: The number of best results to store.
      batch_features: The current suggested features batch array with a
        dimension of (batch_size, feature_dim).
      batch_rewards: The current reward batch array with a dimension of
        (batch_size,).

    Returns:
      trials:
    """
    all_rewards = jnp.concatenate([batch_rewards, best_results.rewards], axis=0)
    all_features = jnp.concatenate(
        [batch_features, best_results.features], axis=0
    )
    top_indices = jnp.argpartition(-all_rewards, count - 1)[:count]
    return VectorizedStrategyResults(
        rewards=all_rewards[top_indices],
        features=all_features[top_indices],
    )


def best_candidates_to_trials(
    best_results: VectorizedStrategyResults,
    converter: ArrayConverter,
) -> list[vz.Trial]:
  """Returns the best candidate trials in the original search space."""
  trials = []
  sorted_ind = jnp.argsort(-best_results.rewards)
  for i in range(len(best_results.rewards)):
    # Create trials and convert the strategy features back to parameters.
    ind = sorted_ind[i]
    trial = vz.Trial(
        parameters=converter.to_parameters(
            jnp.expand_dims(best_results.features[ind], axis=0)
        )[0]
    )
    trial.complete(vz.Measurement({'acquisition': best_results.rewards[ind]}))
    trials.append(trial)
  return trials


def trials_to_sorted_array(
    prior_trials: list[vz.Trial],
    converter: ArrayConverter,
) -> Optional[types.Array]:
  """Sorts trials by the order they were created and converts to array."""
  if prior_trials:
    prior_trials = sorted(prior_trials, key=lambda x: x.creation_time)
    prior_features = converter.to_features(prior_trials)
    # TODO: Update this code to work more cleanly with
    # PaddedArrays.
    if isinstance(converter, converters.PaddedTrialToArrayConverter):
      # We need to mask out the `NaN` padded trials with zeroes.
      prior_features = prior_features.padded_array
      prior_features[len(prior_trials) :, ...] = 0.0
  else:
    prior_features = None
  return prior_features


@attr.define
class VectorizedOptimizerFactory:
  """Vectorized strategy optimizer factory."""

  strategy_factory: VectorizedStrategyFactory
  max_evaluations: int = 75_000
  suggestion_batch_size: int = 25

  def __call__(
      self,
      converter: ArrayConverter,
  ) -> VectorizedOptimizer:
    """Generates a new VectorizedOptimizer object."""
    strategy = self.strategy_factory(
        converter, suggestion_batch_size=self.suggestion_batch_size
    )
    n_feature_dimensions = sum(
        spec.num_dimensions for spec in converter.output_specs
    )
    n_feature_dimensions_with_padding = converter.to_features([]).shape[-1]
    return VectorizedOptimizer(
        strategy=strategy,
        n_feature_dimensions=n_feature_dimensions,
        n_feature_dimensions_with_padding=n_feature_dimensions_with_padding,
        suggestion_batch_size=self.suggestion_batch_size,
        max_evaluations=self.max_evaluations,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Quasi-random designer."""

import collections
import math
import sys
import time
from typing import Iterable, List, Optional, Sequence

import attr
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.interfaces import serializable
from vizier.pyvizier import converters


################################################################################
# PyTypes
NumpyArraySpec = converters.NumpyArraySpec
NumpyArraySpecType = converters.NumpyArraySpecType
ScaleType = vz.ScaleType
################################################################################


def _is_prime(n: int) -> bool:
  """Check if `n` is a prime number."""
  return all(n % i != 0 for i in range(2, int(n**0.5) + 1)) and n != 2


def _generate_primes(n: int) -> List[int]:
  """Generate all primes less than `n` (except 2) using the Sieve of Sundaram."""
  if n <= 2:
    raise ValueError('n has to be larger than 2: %s' % n)
  half_m1 = int((n - 2) / 2)
  sieve = [0] * (half_m1 + 1)
  for outer in range(1, half_m1 + 1):
    inner = outer
    while outer + inner + 2 * outer * inner <= half_m1:
      sieve[outer + inner + (2 * outer * inner)] = 1
      inner += 1
  return [2 * i + 1 for i in range(1, half_m1 + 1) if sieve[i] == 0]


def _init_primes(num_dimensions: int) -> List[int]:
  """Get list of primes for all parameters in the search space."""
  primes = []
  prime_attempts = 1
  while len(primes) < num_dimensions + 1:
    primes = _generate_primes(1000 * prime_attempts)
    prime_attempts += 1
  primes = primes[-num_dimensions - 1 : -1]
  return primes


@attr.define(init=False, kw_only=True)
class _HaltonSequence(serializable.PartiallySerializable):
  """Encapsulates the generation of a scrambled halton sequence.

  Specifically, this class is a fork inspired by the implementation of
  scrambled Halton sequence of quasi-random numbers (by Google):

  https://github.com/google/uncertainty-baselines/blob/main/uncertainty_baselines/halton.py
  """

  _num_dimensions: int = attr.field(validator=attr.validators.instance_of(int))

  _skip_points: int = attr.field(validator=attr.validators.instance_of(int))

  _num_points_generated: int = attr.field(
      validator=attr.validators.instance_of(int)
  )

  _primes: List[int] = attr.field(
      init=True,
      converter=list,
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(int),
          iterable_validator=attr.validators.instance_of(Iterable),
      ),
  )

  _scramble: bool = attr.field(
      default=True, validator=attr.validators.instance_of(bool), kw_only=True
  )
  _seed: int = attr.field(default=0)

  def __init__(
      self,
      num_dimensions: int,
      *,
      skip_points: int,
      num_points_generated: int = 0,
      primes_override: Optional[List[int]] = None,
      scramble: bool = True,
      seed: int = 0,
  ):
    """Create a Halton sequence generator.

    Args:
      num_dimensions: Number of dimensions for each point in the seqeunce. This
        corresponds to the number of parameters in the Vizier search space.
      skip_points: The number of initial points that should be skipped before
        the first point is returned.
      num_points_generated: Number of points that have already been generated.
      primes_override: If supplied, use these primes to seed each dimension of
        the Halton sequence. This is useful for testing. NOTE: These values are
        not validated, so it is the responsibility of the user to supply
        legitimate primes.
      scramble: If True, will scramble the resulting Halton sequence. Set this
        to False for testing.
      seed: random seed

    Returns:
      A HaltonSequence object.
    """
    if skip_points < 0:
      raise ValueError('skip_points must be non-negative: %s' % skip_points)

    if primes_override:
      if len(primes_override) != num_dimensions:
        raise ValueError(
            'Expected len(primes_overrides) and num_dimensions to '
            f'be the same size. len(primes_overrides): {len(primes_override)},'
            f'num_dimensions: {num_dimensions}'
        )
      primes = primes_override
    else:
      primes = _init_primes(num_dimensions)

    self.__attrs_init__(
        num_dimensions=num_dimensions,
        skip_points=skip_points,
        num_points_generated=num_points_generated,
        primes=primes,
        scramble=scramble,
        seed=seed,
    )

  def load(self, metadata: vz.Metadata) -> None:
    self._num_points_generated = int(
        metadata.ns('halton')['num_points_generated']
    )
    self._seed = int(metadata.ns('halton')['permutation_seed'])

  def dump(self) -> vz.Metadata:
    metadata = vz.Metadata()
    metadata.ns('halton')['num_points_generated'] = str(
        self._num_points_generated
    )
    metadata.ns('halton')['permutation_seed'] = str(self._seed)
    return metadata

  def _get_scrambled_halton_value(self, index: int, base: int) -> float:
    """Get a scrambled Halton value for a given `index` using `base`.

    For example, if `index` can be represented in `base` b as:

    `index = 5 * b^{2} + 2 * b^{1} + 3`

    then the Halton value will be:

    `Halton(index) = 5 * b^{-1} + 2 * b^{-2} + 3 * b^{-3}.`

    If we choose to scramble, then the coefficients (5,2,3) will be randomly
    permuted seeded by self._seed and base, leading to a
    roughly uniform distribution of Halton values over the interval [0,1].

    Args:
      index: Index to be converted to `base`.
      base: Base used for conversion.

    Returns:
      (Possibly scrambled) Halton value.
    """
    if not _is_prime(base):
      raise ValueError('base is not prime: %s' % base)

    result = 0.0
    base_rec = 1.0 / base
    f = base_rec
    i = index + 1  # For index 0 we want 1/base returned, not 0.

    # Use a fixed seed to generate the permutation in a deterministic way.
    if self._scramble:
      permutation = list(range(1, base))
      rng = np.random.RandomState((self._seed, base))
      rng.shuffle(permutation)
      permutation = [0] + permutation
    while i > 0:
      i, mod = divmod(i, base)
      if self._scramble:
        result += f * permutation[mod]
      else:
        result += f * mod
      f *= base_rec

    if (0.0 > result) or (result > 1.0):
      raise ValueError(
          'Something wrong has happened; halton_value should be within [0,'
          ' 1]: %f' % result
      )
    return result

  def get_next_list(self) -> List[float]:
    """Get the next list in a sequence seeded by `primes`.

    This implementation and its associated unit tests are inspired by another
    implementation from Google.

    https://github.com/mlcommons/algorithmic-efficiency/blob/main/algorithmic_efficiency/halton.py

    Returns:
      An sublist of the Halton sequence. Every value in the list should be
      within [0,1].
    """
    index = self._num_points_generated + self._skip_points
    halton_list = [
        self._get_scrambled_halton_value(index, prime) for prime in self._primes
    ]
    self._num_points_generated += 1
    return halton_list


class QuasiRandomDesigner(vza.PartiallySerializableDesigner):
  """Sample points using quasi-random search from the scaled search space.

  This implementation uses a scrambled Halton sequence. The core idea, is given
  a prime `p`, we can map an index `n` to its Halton value by converting `n` to
  base `p`, e.g.:

  `n = c3 * p^{3} + c2 * p^{2} + c1 * p^{1} + c0`

  and then perform an inversion on the base to obtain:

  `halton(n) = c3 * p^{-1} + c2 * p^{-2} + c1 * p^{-3} + c0 * p^{-4}`

  To obtain a pseuo-uniform distribution, we permute the coefficients:

  `permuted_halton(n) = c2 * p^{-1} + c1 * p^{-2} + c0 * p^{-3} + c3 * p^{-4}`
  """

  def __init__(
      self,
      search_space: vz.SearchSpace,
      *,
      skip_points: int = 1000,
      seed: Optional[int] = None,
  ):
    """Init.

    Args:
      search_space: Must be a flat search space.
      skip_points: If positive, then these first points in the sequence are
        discarded in order to avoid unwanted correlations.
      seed: Random seed.
    """
    if seed is None:
      seed = np.int32(time.time())
    if search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.'
      )

    def create_input_converter(pc):
      return converters.DefaultModelInputConverter(
          pc,
          scale=True,
          max_discrete_indices=sys.maxsize,
          float_dtype=np.float64,
      )

    self._converter = converters.DefaultTrialConverter(
        [create_input_converter(pc) for pc in search_space.parameters]
    )

    for spec in self._converter.output_specs.values():
      if spec.type not in [
          NumpyArraySpecType.CONTINUOUS,
          NumpyArraySpecType.DISCRETE,
      ]:
        raise ValueError(f'Unsupported type: {spec.type} in {spec}')
      if spec.num_dimensions != 1:
        raise ValueError(
            'Multi-dimensional discrete types are unsuppored. Received spec: %s'
            % spec
        )

    self._halton_generator = _HaltonSequence(
        len(search_space.parameters),
        skip_points=skip_points,
        num_points_generated=0,
        scramble=True,
        seed=seed,
    )

    self._output_specs = tuple(self._converter.output_specs.values())

  @classmethod
  def from_problem(
      cls, problem: vz.ProblemStatement, seed: Optional[int] = None
  ):
    """For wrapping via `PartiallySerializableDesignerPolicy`."""
    return QuasiRandomDesigner(problem.search_space, seed=seed)

  def load(self, metadata: vz.Metadata) -> None:
    self._halton_generator.load(metadata)

  def dump(self) -> vz.Metadata:
    return self._halton_generator.dump()

  def _generate_discrete_point(
      self, spec: NumpyArraySpec, halton_value: float
  ) -> int:
    """Generate a discrete parameter value from a Halton value."""

    # +1 because the bounds are inclusive on both ends.
    num_discrete_options = spec.bounds[1] - spec.bounds[0] + 1 - spec.num_oovs
    # Get a number in [0,  num_discrete_options].
    halton_value *= num_discrete_options
    # Get an integer between 0 and num_discrete_options-1 (inclusive).
    halton_value = int(math.floor(halton_value))
    return halton_value + int(spec.bounds[0])

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    pass

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Suggest new suggestions, taking into account `count`."""
    count = count or 1

    sample = collections.defaultdict(list)
    for _ in range(count):
      # Dimension of halton_list is [P], where P is number of primes used.
      halton_list = self._halton_generator.get_next_list()
      for dimension_index, spec in enumerate(self._output_specs):
        # Only NumpyArraySpecType.CONTINUOUS and NumpyArraySpecType.DISCRETE
        # are supported.
        halton_value = halton_list[dimension_index]
        if spec.type == NumpyArraySpecType.CONTINUOUS:
          # Trial-Numpy converter was configured to scale values to [0, 1].
          # We sample from that range and rely on it scaled correctly when the
          # Trials are created.
          # halton_value is also within [0, 1].
          sample[spec.name].append(np.float64(halton_value))
        elif spec.type == NumpyArraySpecType.DISCRETE:
          # Trial-Numpy converter expects an integer for discrete/categorical
          # parameters.
          sample[spec.name].append(
              np.int64(self._generate_discrete_point(spec, halton_value))
          )
        else:
          raise ValueError(
              f'Unsupported spec type: {spec.type}. self._converter should be'
              ' configured to return CONTINUOUS or DISCRETE specs only.'
          )
    sample = {
        name: np.expand_dims(np.asarray(elements), axis=-1)
        for (name, elements) in sample.items()
    }
    return [
        vz.TrialSuggestion(p) for p in self._converter.to_parameters(sample)
    ]



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utility functions for performance tracking.

A typical use case is:

class MyDesigner:

  def _compute(self, x: int):
    pass

  @profiler.record_runtime
  def suggest(self) -> None:
    # An example of how to annotate an inline function call, so the function
    # name `_compute` is used. Note that the name must be specified.
    with profiler.record_runtime_context(name='compute'):
      self._compute(x=1)

    # An example of how to annotate an inline code block:
    with profiler.record_runtime_context(name='compute_twice'):
      self._compute(x=2)
      self._compute(x=2)

  @profiler.record_runtime(name_prefix='designer')
  def suggest2(self) -> None:
    pass

  @profiler.record_runtime(name_prefix='designer', name='suggest')
  @jax.jit
  def suggest3(self) -> None:
    pass


designer = MyDesigner()
designer.suggest()

profiler.Storage().runtimes() returns a dict with these keys:
* 'suggest' (the name of the suggest method)
* 'compute' (from the call with x=1)
* 'compute_twice' (from the calls with x=2)
* 'designer.suggest2' (the name of the suggest2 method with the prefix)
* 'designer.suggest' (the specified decorator name for suggest3 with the prefix)
values will be lists of datetime.timedelta objects.

profiler.Storage().runtimes('suggest') will return a list with
datetime.timedelta objects for only `suggest` function calls.

NOTE that @profiler.record_runtime should appear FIRST, before the
@jax.jit decorator, or only the first runtime will be recorded.
"""
import collections
import contextlib
import datetime
import functools
from typing import Any, Generator

from absl import logging
import jax


class Storage:
  """Storage for function call runtimes."""

  def __new__(cls):
    if not hasattr(cls, '_instance'):
      cls._instance = super(Storage, cls).__new__(cls)
      # Add attributes below this line.
      # -------------------------------
      # A dict to track runtime performance.
      # Keys are computation names, values are computation duration.
      cls._instance.storage = collections.defaultdict(list)
    return cls._instance

  def record(self, name: str, duration: datetime.timedelta) -> None:
    """Records the runtime for the function."""
    self._instance.storage[name].append(duration)

  def reset(self) -> None:
    """Removes all performance information collected."""
    self._instance.storage.clear()

  def runtimes(self) -> dict[str, list[datetime.timedelta]]:
    """Returns the stored runtimes.

    Returns:
      A dict with all function names and their timedelta runtimes.
    """
    return dict(self._instance.storage)


def record_runtime(
    func=None,
    *,
    name_prefix: str = '',
    name: str = '',
    also_log: bool = False,
) -> Any:
  """A decorator for recording the runtime of functions.

  Args:
    func: Function being decorated.
    name_prefix: A prefix to add to the function name.
    name: The name to record. Defaults to func.__name__.
    also_log: Whether to also logging.info the runtime duration.

  Returns:
    The function return value.
  """
  # This is required for the decorator to work both with and without
  # optional arguments.
  if func is None:
    return functools.partial(
        record_runtime,
        name_prefix=name_prefix,
        name=name,
        also_log=also_log,
    )
  name = name or func.__name__
  if name_prefix:
    full_name = f'{name_prefix}.{name}'
  else:
    full_name = name

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    """Calculates runtime of the given function."""
    start = datetime.datetime.now()
    # This works fine for jitted functions as well.
    result = func(*args, **kwargs)
    # If running on an acceleartor, wait for the async dispatch result.
    # See https://jax.readthedocs.io/en/latest/async_dispatch.html
    result = jax.block_until_ready(result)
    duration = datetime.datetime.now() - start
    Storage().record(full_name, duration)
    if also_log:
      logging.info(
          '%s: Duration %s seconds', full_name, duration.total_seconds()
      )
    return result

  return wrapper


@contextlib.contextmanager
def record_runtime_context(
    name: str, also_log: bool = False
) -> Generator[None, None, None]:
  """A context manager to record the runtime of a function call/code block.

  Args:
    name: The name of the function call or code block to record timing for
    also_log: Whether to also logging.info the runtime duration.

  Yields:
    Nothing.
  """
  start = datetime.datetime.now()
  yield
  duration = datetime.datetime.now() - start
  Storage().record(name, duration)
  if also_log:
    logging.info('%s: Duration %s seconds', name, duration.total_seconds())



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for convergence_curve."""

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.analyzers import convergence_curve as convergence

from absl.testing import absltest
from absl.testing import parameterized


def _gen_trials(values):
  """Returns trials where trials[i] has empty metric name equal to values[i]."""
  trials = []
  for v in values:
    trial = pyvizier.Trial()
    trials.append(
        trial.complete(
            pyvizier.Measurement(metrics={'': pyvizier.Metric(value=v)})))
  return trials


class ConvergenceCurveTest(absltest.TestCase):

  def test_align_xs_merge_ys_on_different_lengths(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 2, 3]),
        ys=np.array([[2, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    c2 = convergence.ConvergenceCurve(
        xs=np.array([1]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    aligned = convergence.ConvergenceCurve.align_xs([c1, c2])[0]

    np.testing.assert_array_equal(aligned.xs, [1, 2, 3])
    np.testing.assert_array_equal(aligned.ys,
                                  np.array([[2, 1, 1], [3, np.nan, np.nan]]))

  def test_align_xs_merge_ys_on_distinct_xvalues(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 3, 4]),
        ys=np.array([[2, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    c2 = convergence.ConvergenceCurve(
        xs=np.array([2]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    aligned = convergence.ConvergenceCurve.align_xs([c1, c2])[0]

    np.testing.assert_array_equal(aligned.xs.shape, (3,))
    np.testing.assert_array_equal(aligned.ys,
                                  np.array([[2, 1.25, 1], [3, np.nan, np.nan]]))

  def test_align_xs_merge_ys_with_interpolation(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 2, 3, 4, 5]),
        ys=np.array([[2, 2, 1, 0.5, 0.5], [1, 1, 1, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    aligned = convergence.ConvergenceCurve.align_xs(
        [c1], interpolate_repeats=True
    )[0]

    np.testing.assert_array_equal(aligned.xs, np.array([1, 2, 3, 4, 5]))
    np.testing.assert_array_equal(
        aligned.ys, np.array([[2, 1.5, 1.0, 0.5, 0.5], [1, 1, 1, 1, 1]]))

  def test_extrapolate_ys_with_steps(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 2, 3, 4]),
        ys=np.array([[2, 1.5, 1, 0.5], [1, 1, 1, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)

    extra_c1 = convergence.ConvergenceCurve.extrapolate_ys(c1, steps=2)

    np.testing.assert_array_equal(extra_c1.xs.shape, (6,))
    np.testing.assert_array_equal(
        extra_c1.ys,
        np.array([[2, 1.5, 1.0, 0.5, 0.0, -0.5], [1, 1, 1, 1, 1, 1]]))

  def test_align_xs_merge_ys_on_increasing_and_dicreasing_fails(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 3, 4]),
        ys=np.array([[2, 3, 3]]),
        trend=convergence.ConvergenceCurve.YTrend.INCREASING,
    )
    c2 = convergence.ConvergenceCurve(
        xs=np.array([2]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    with self.assertRaisesRegex(ValueError, 'increasing'):
      # pylint: disable=[expression-not-assigned]
      convergence.ConvergenceCurve.align_xs([c1, c2])[0]

  def test_align_xs_keep_ys(self):
    c1 = convergence.ConvergenceCurve(
        xs=np.array([1, 3, 4]),
        ys=np.array([[2, 1, 1], [8, 3, 1]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )
    c2 = convergence.ConvergenceCurve(
        xs=np.array([2]),
        ys=np.array([[3]]),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING,
    )

    aligned1, aligned2 = convergence.ConvergenceCurve.align_xs(
        [c1, c2], keep_curves_separate=True
    )

    np.testing.assert_array_equal(aligned1.xs, np.array([1.0, 2.5, 4.0]))
    np.testing.assert_array_equal(aligned2.xs, np.array([1.0, 2.5, 4.0]))

    np.testing.assert_array_equal(
        aligned1.ys, np.array([[2.0, 1.25, 1.0], [8.0, 4.25, 1.0]])
    )
    np.testing.assert_array_equal(
        aligned2.ys, np.array([[3.0, np.nan, np.nan]])
    )


class ConvergenceCurveConverterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('maximize', pyvizier.ObjectiveMetricGoal.MAXIMIZE, [[2, 2, 3]]),
      ('minimize', pyvizier.ObjectiveMetricGoal.MINIMIZE, [[2, 1, 1]]))
  def test_convert_basic(self, goal, expected):
    trials = _gen_trials([2, 1, 3])
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal))
    curve = generator.convert(trials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_equal(curve.ys, expected)

  @parameterized.named_parameters(
      ('maximize', pyvizier.ObjectiveMetricGoal.MAXIMIZE, [[2, 2, 3]]),
      ('minimize', pyvizier.ObjectiveMetricGoal.MINIMIZE, [[-2, -1, -1]]),
  )
  def test_convert_flip_signs(self, goal, expected):
    trials = _gen_trials([2, 1, 3])
    generator = convergence.ConvergenceCurveConverter(
        pyvizier.MetricInformation(name='', goal=goal), flip_signs_for_min=True
    )
    curve = generator.convert(trials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_equal(curve.ys, expected)


class HyperConvergenceCurveConverterTest(parameterized.TestCase):

  def test_convert_basic(self):
    generator = convergence.HypervolumeCurveConverter([
        pyvizier.MetricInformation(
            name='max', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
        ),
        pyvizier.MetricInformation(
            name='min', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE
        ),
    ])
    pytrials = []
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': 2.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 3.0, 'min': -1.0})
        )
    )
    pytrials.append(
        pyvizier.Trial().complete(
            pyvizier.Measurement(metrics={'max': 4.0, 'min': -2.0})
        )
    )

    curve = generator.convert(pytrials)
    np.testing.assert_array_equal(curve.xs, [1, 2, 3])
    np.testing.assert_array_almost_equal(curve.ys, [[0.0, 3.0, 8.0]], decimal=1)


class LogEfficiencyConvergenceComparatorTest(absltest.TestCase):

  def setUp(self):
    super(LogEfficiencyConvergenceComparatorTest, self).setUp()
    xs = np.array(range(0, 20))
    xs_t = xs.reshape(1, len(xs))
    self._baseline_curve = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-0.9, -1.0, -1.1]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)

    self._worse_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(-0.5 * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    self._better_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-1.5, -1.8, -2.0]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)

  def test_get_relative_efficiency_curve(self):
    baseline_length = len(self._baseline_curve.xs)
    rel_effiency = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).log_efficiency_curve()
    higher_quantile = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve,
        compared_curve=self._better_curves,
        compared_quantile=0.9,
    ).log_efficiency_curve()

    self.assertEqual(rel_effiency.ys.shape, (1, baseline_length))
    self.assertEqual(higher_quantile.ys.shape, (1, baseline_length))
    # Better curves should have positive efficiency.
    self.assertTrue((rel_effiency.ys >= 0.0).all())
    # Higher quantile means better efficiency which means more positive scores.
    self.assertTrue((higher_quantile.ys >= rel_effiency.ys).all())

  def test_get_relative_efficiency_flat(self):
    flat_curve = convergence.ConvergenceCurve(
        xs=np.array(range(0, 20)),
        ys=np.array([4.0, 3.0, 2.0] + [1.5] * 17).reshape(1, 20),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    self_eff = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=flat_curve, compared_curve=flat_curve
    ).log_efficiency_curve()
    # Relative efficiency of a curve on itself is close to 0.
    self.assertAlmostEqual(np.linalg.norm(self_eff.ys), 0.0, delta=0.1)

  def test_get_relative_efficiency_short_curve(self):
    baseline_length = len(self._baseline_curve.xs)
    short_length = round(baseline_length / 2)
    short_curve = convergence.ConvergenceCurve(
        xs=self._baseline_curve.xs[:short_length],
        ys=self._baseline_curve.ys[:, :short_length],
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    short_efficiency = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=short_curve
    ).log_efficiency_curve()
    self.assertEqual(short_efficiency.ys.shape, (1, baseline_length))
    self.assertEqual(float(short_efficiency.ys[:, -1]), -float('inf'))

  def test_get_efficiency_score(self):
    # Higher compared quantile should increase score. Higher baseline
    # quantile should decrease score.
    median_score = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).score()
    self.assertGreater(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve,
            compared_curve=self._better_curves,
            compared_quantile=0.9,
        ).score(),
        median_score,
    )
    self.assertLess(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve,
            compared_curve=self._better_curves,
            baseline_quantile=0.9,
        ).score(),
        median_score,
    )

  def test_effiency_score_symmetry(self):
    base_score = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._baseline_curve, compared_curve=self._better_curves
    ).score()
    reversed_score = convergence.LogEfficiencyConvergenceCurveComparator(
        baseline_curve=self._better_curves, compared_curve=self._baseline_curve
    ).score()
    self.assertAlmostEqual(base_score, -reversed_score, delta=0.01)

  def test_efficiency_score_value(self):
    xs = self._baseline_curve.xs
    xs_t = xs.reshape(1, len(xs))

    worse_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(-0.5 * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    better_curves = convergence.ConvergenceCurve(
        xs=xs,
        ys=np.exp(np.array([-1.5, -1.8, -2.0]).reshape(3, 1) * xs_t),
        trend=convergence.ConvergenceCurve.YTrend.DECREASING)
    # Efficiency score for exponential curves can be approximated.

    self.assertGreater(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve, compared_curve=better_curves
        ).score(),
        0.4,
    )
    self.assertLess(
        convergence.LogEfficiencyConvergenceCurveComparator(
            baseline_curve=self._baseline_curve, compared_curve=worse_curves
        ).score(),
        -0.4,
    )

  def test_comparator_failure(self):
    unknown_curve = convergence.ConvergenceCurve(
        xs=self._baseline_curve.xs, ys=self._baseline_curve.ys)
    with self.assertRaisesRegex(ValueError, 'increasing or decreasing'):
      convergence.LogEfficiencyConvergenceCurveComparator(
          baseline_curve=unknown_curve, compared_curve=self._baseline_curve
      )


class SimpleConvergenceComparatorTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'ys1': np.array([[11, 12, 20, 50, 100, 300]]),
          'ys2': np.array([[1, 2, 10]]),
          'cutoff': None,
          'res': 0.0,
      },
      {
          'ys1': np.array([[1, 2, 10]]),
          'ys2': np.array([[11, 12, 20, 50, 100, 300]]),
          'cutoff': None,
          'res': 1.0,
      },
      {
          'ys1': np.array([[1, 4, 8, 10]]),
          'ys2': np.array([[2, 5, 6, 8]]),
          'cutoff': None,
          'res': 0.5,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'cutoff': 1,
          'res': 0.25,
      },
      {
          'ys1': np.array([[1, 4, 8, 10, 12]]),
          'ys2': np.array([[2, 5, 6, 8, 10]]),
          'cutoff': 3,
          'res': 0.0,
      },
      {
          'ys1': np.array([[2, 5, 6, 8, 10]]),
          'ys2': np.array([[1, 4, 8, 10, 12]]),
          'cutoff': 3,
          'res': 1.0,
      },
  )
  def test_score_one_curve_above_other(self, ys1, ys2, res, cutoff):
    xs1 = np.arange(ys1.shape[1])
    xs2 = np.arange(ys2.shape[1])
    curve1 = convergence.ConvergenceCurve(
        xs=xs1, ys=ys1, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    curve2 = convergence.ConvergenceCurve(
        xs=xs2, ys=ys2, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    comparator = convergence.SimpleConvergenceCurveComparator(
        curve1, curve2, xs_cutoff=cutoff
    )
    self.assertEqual(comparator.score(), res)


class PercentageBetterConvergenceComparatorTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'ys1': np.array([[1, 2, 3, 4], [1, 2, 3, 4]]),
          'ys2': np.array([[-1, 20, 30, 70]]),
          'res': 0.5,
      },
      {
          'ys1': np.array([[1, 2, 3, 4]]),
          'ys2': np.array([[10, 20, 30, 70]]),
          'res': 1.0,
      },
      {
          'ys1': np.array([[10, 20, 30, 70]]),
          'ys2': np.array([[1, 2, 3, 4]]),
          'res': -1.0,
      },
  )
  def test_score(self, ys1, ys2, res):
    xs1 = np.arange(ys1.shape[1])
    xs2 = np.arange(ys2.shape[1])
    curve1 = convergence.ConvergenceCurve(
        xs=xs1, ys=ys1, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    curve2 = convergence.ConvergenceCurve(
        xs=xs2, ys=ys2, trend=convergence.ConvergenceCurve.YTrend.INCREASING
    )
    comparator = convergence.PercentageBetterConvergenceCurveComparator(
        curve1, curve2
    )
    self.assertEqual(comparator.score(), res)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Converter for padding trials."""

from typing import Sequence, Tuple

import numpy as np

from vizier import pyvizier
from vizier.pyvizier.converters import core
from vizier.pyvizier.converters import padding


class PaddedTrialToArrayConverter:
  """Converts trials to arrays and pads / masks them."""

  _experimental_override = 'I am aware that this code may break at any point.'

  def __init__(
      self,
      impl: core.TrialToArrayConverter,
      experimental_override: str = '',
      *,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(
          num_trials=padding.PaddingType.NONE,
          num_features=padding.PaddingType.NONE,
      ),
  ):
    """SHOULD NOT BE USED! Use factory classmethods e.g. from_study_config."""

    if experimental_override != self._experimental_override:
      raise ValueError(
          'Set "experimental_override" if you want to call __init__ directly. '
          'Otherwise, use TrialToArrayConverter.from_study_config.'
      )
    self._impl = impl
    self._padding_schedule = padding_schedule

  def to_features(
      self, trials: Sequence[pyvizier.Trial]
  ) -> padding.PaddedArray:
    """Returns the features array with dimension: (n_trials, n_features)."""
    # Pad up the features.
    features = self._impl.to_features(trials)
    return padding.pad_features(features, self._padding_schedule)

  def to_labels(self, trials: Sequence[pyvizier.Trial]) -> padding.PaddedArray:
    """Returns the labels array with dimenion: (n_trials, n_metrics)."""
    # Pad up the labels.
    labels = self._impl.to_labels(trials)
    return padding.pad_labels(labels, self._padding_schedule)

  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[padding.PaddedArray, padding.PaddedArray]:
    return self.to_features(trials), self.to_labels(trials)

  def to_parameters(self, arr: np.ndarray) -> Sequence[pyvizier.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    # Undo the padding of the parameters in this array.
    if self._padding_schedule is not None:
      # Get the original shape for the features without padding and
      # remove them.
      n_features = sum([spec.num_dimensions for spec in self.output_specs])
      arr = arr[..., :n_features]
    return self._impl.to_parameters(arr)

  @classmethod
  def from_study_config(
      cls,
      study_config: pyvizier.ProblemStatement,
      *,
      scale: bool = True,
      padding_schedule: padding.PaddingSchedule = padding.PaddingSchedule(
          num_trials=padding.PaddingType.NONE,
          num_features=padding.PaddingType.NONE,
      ),
      pad_oovs: bool = True,
      max_discrete_indices: int = 0,
      flip_sign_for_minimization_metrics: bool = True,
      dtype=np.float64,
  ) -> 'PaddedTrialToArrayConverter':
    """From study config.

    Args:
      study_config:
      scale: If True, scales the parameters to [0, 1] range.
      padding_schedule: Pads features and labels according to a padding
        schedule. This is to reduce the number of shapes a designer may see, and
        thus reduce JIT retracing.
      pad_oovs: If True, add an extra dimension for out-of-vocabulary values for
        non-CONTINUOUS parameters.
      max_discrete_indices: For DISCRETE and INTEGER parameters that have more
        than this many feasible values will be continuified. When generating
        suggestions, values are rounded to the nearest feasible value. Note this
        default is different from the default in DefaultModelInputConverter.
      flip_sign_for_minimization_metrics: If True, flips the metric signs so
        that every metric maximizes.
      dtype: dtype

    Returns:
      PaddedTrialToArrayConverter
    """
    converter = core.TrialToArrayConverter.from_study_config(
        study_config,
        scale=scale,
        pad_oovs=pad_oovs,
        max_discrete_indices=max_discrete_indices,
        flip_sign_for_minimization_metrics=flip_sign_for_minimization_metrics,
        dtype=dtype,
    )
    return cls(
        converter, cls._experimental_override, padding_schedule=padding_schedule
    )

  @property
  def output_specs(self) -> Sequence[core.NumpyArraySpec]:
    return self._impl.output_specs

  @property
  def metric_specs(self) -> Sequence[pyvizier.MetricInformation]:
    return self._impl.metric_specs

  @property
  def padding_schedule(self) -> padding.PaddingSchedule:
    return self._padding_schedule



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Running Ray Tuners: https://docs.ray.io/en/latest/ray-air/tuner.html."""

from typing import Any, Callable, List, Optional, Tuple

import numpy as np
from ray import air
from ray import data
from ray import tune
from vizier import pyvizier
from vizier._src.raytune import converters
from vizier.benchmarks import experimenters


# See https://docs.ray.io/en/latest/data/dataset.html
def run_tune_distributed(
    run_tune_args_list: List[Tuple[Any]],
    run_tune: Callable[[Any], tune.result_grid.ResultGrid],
) -> List[tune.result_grid.ResultGrid]:
  """Distributes tuning via Ray datasets API for MapReduce purposes.

  NOTE: There are no datasets processed. However, all MapReduce operations
  are now done in the Datasets API under Ray.

  Args:
    run_tune_args_list: List of Tuples that are to be passed into run_tune.
    run_tune: Callable that accepts args from previous list.

  Returns:
    List of results.
  """
  ds = data.from_items(run_tune_args_list)
  ds = ds.map(lambda args_tuple: run_tune(*args_tuple))
  return ds.take_all()


def run_tune_bbob(
    function_name: str,
    dimension: int,
    shift: Optional[np.ndarray] = None,
    tune_config: Optional[tune.TuneConfig] = None,
    run_config: Optional[air.RunConfig] = None,
) -> tune.result_grid.ResultGrid:
  """Runs Ray Tuners for BBOB problems.

  See https://docs.ray.io/en/latest/tune/key-concepts.html
  For more information on Tune and Run configs, see
  https://docs.ray.io/en/latest/ray-air/tuner.html

  Args:
    function_name: BBOB function name.
    dimension: Dimension of BBOB function.
    shift: Shift of BBOB function.
    tune_config: Ray Tune Config.
    run_config: Ray Run Config.

  Returns:
  """
  experimenter_factory = experimenters.BBOBExperimenterFactory(
      name=function_name, dim=dimension
  )
  if shift is not None:
    experimenter_factory = experimenters.SingleObjectiveExperimenterFactory(
        base_factory=experimenter_factory, shift=shift
    )
  experimenter = experimenter_factory()
  problem = experimenter.problem_statement()

  param_space = converters.SearchSpaceConverter.to_dict(problem.search_space)
  objective = converters.ExperimenterConverter.to_callable(experimenter)

  metric_info = problem.metric_information.item()
  if tune_config is None:
    tune_config = tune.TuneConfig()
  tune_config.metric = metric_info.name
  if metric_info.goal == pyvizier.ObjectiveMetricGoal.MINIMIZE:
    tune_config.mode = 'min'
  else:
    tune_config.mode = 'max'

  def objective_fn(config: pyvizier.ParameterDict) -> None:
    # Config contains parameter names to values and is autopopulated for each
    # Trial. Evaluation is static for BBOB so we simply loop.
    for _ in range(tune_config.num_samples):
      result_dict = objective(config)
      air.session.report(result_dict)

  tuner = tune.Tuner(
      objective_fn,
      param_space=param_space,
      run_config=run_config,
      tune_config=tune_config,
  )
  return tuner.fit()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for acquisitions."""

from unittest import mock

import jax
from jax import numpy as jnp
from jax.config import config
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.jax import types
from vizier.pyvizier import converters

from absl.testing import absltest


tfd = tfp.distributions


def _build_mock_continuous_array_specs(n):
  continuous_spec = mock.create_autospec(converters.NumpyArraySpec)
  continuous_spec.type = converters.NumpyArraySpecType.CONTINUOUS
  continuous_spec.num_dimensions = 1
  return [continuous_spec] * n


class AcquisitionsTest(absltest.TestCase):

  def test_ucb(self):
    acq = acquisitions.UCB(coefficient=2.0)
    self.assertAlmostEqual(acq(tfd.Normal(0.1, 1)), 2.1)

  def test_hvs(self):
    acq = acquisitions.HyperVolumeScalarization(coefficient=2.0)
    self.assertAlmostEqual(acq(tfd.Normal(0.1, 1)), 0.1)

  def test_ei(self):
    acq = acquisitions.EI()
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1), labels=jnp.array([0.2])),
        0.34635347,
    )

  def test_pi(self):
    acq = acquisitions.PI()
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1), labels=jnp.array([0.2])),
        0.46017216,
    )

  def test_qei(self):
    acq = acquisitions.QEI(num_samples=2000)
    batch_shape = [6]
    dist = tfd.Normal(loc=0.1 * jnp.ones(batch_shape), scale=1.0)
    qei = acq(dist, labels=jnp.array([0.2]), seed=jax.random.PRNGKey(0))
    # QEI reduces over the batch shape.
    self.assertEmpty(qei.shape)

    dist_single_point = tfd.Normal(jnp.array([0.1], dtype=jnp.float64), 1)
    qei_single_point = acq(
        dist_single_point, labels=jnp.array([0.2]), seed=jax.random.PRNGKey(0)
    )
    # Parallel matches non-parallel for a single point.
    np.testing.assert_allclose(qei_single_point, 0.346, atol=1e-2)
    self.assertEmpty(qei_single_point.shape)

  def test_qucb_shape(self):
    acq = acquisitions.QUCB()
    batch_shape = [6]
    dist = tfd.Normal(loc=0.1 * jnp.ones(batch_shape), scale=1.0)
    qucb = acq(dist, labels=jnp.array([0.2]), seed=jax.random.PRNGKey(0))
    # QUCB reduces over the batch shape.
    self.assertEmpty(qucb.shape)

  def test_qucb_equals_ucb(self):
    # The QUCB coefficient should be multiplied by sqrt(pi/2) for equivalency
    # with the UCB coefficient (assuming a Gaussian distribution).
    acq_ucb = acquisitions.UCB(coefficient=0.5)
    acq_qucb = acquisitions.QUCB(
        num_samples=5000, coefficient=0.5 * np.sqrt(np.pi / 2.0)
    )
    dist_single_point = tfd.Normal(jnp.array([0.1], dtype=jnp.float64), 1)
    qucb_single_point = acq_qucb(
        dist_single_point, labels=jnp.array([0.2]), seed=jax.random.PRNGKey(1)
    )
    ucb_single_point = acq_ucb(dist_single_point, labels=jnp.array([0.2]))
    # Parallel matches non-parallel for a single point.
    np.testing.assert_allclose(
        qucb_single_point, ucb_single_point[0], atol=2e-2
    )
    self.assertEmpty(qucb_single_point.shape)

  def test_multi_acquisition(self):
    ucb = acquisitions.UCB()
    ei = acquisitions.EI()
    acq = acquisitions.MultiAcquisitionFunction({'ucb': ucb, 'ei': ei})
    dist = tfd.Normal(jnp.float64(0.1), 1)
    labels = jnp.array([0.2])
    acq_val = acq(dist, labels=labels)
    ucb_val = ucb(dist, labels=labels)
    ei_val = ei(dist, labels=labels)
    np.testing.assert_allclose(acq_val, jnp.stack([ucb_val, ei_val]))


class TrustRegionTest(absltest.TestCase):

  def test_trust_region_small(self):
    data = types.StochasticProcessModelData(
        features=np.array([
            [0.0, 0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0],
        ]),
        labels=np.array([0.0, 0.0]),
    )
    tr = acquisitions.TrustRegion.build(
        _build_mock_continuous_array_specs(4), data=data
    )

    np.testing.assert_allclose(
        tr.min_linf_distance(
            np.array([
                [0.0, 0.2, 0.3, 0.0],
                [0.9, 0.8, 0.9, 0.9],
                [1.0, 1.0, 1.0, 1.0],
            ]),
        ),
        np.array([0.3, 0.2, 0.0]),
    )
    self.assertAlmostEqual(tr.trust_radius, 0.224, places=3)

  def test_trust_region_bigger(self):
    features = np.vstack(
        [
            [0.0, 0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0],
        ]
        * 10
    )
    labels = np.sum(features, axis=-1)
    data = types.StochasticProcessModelData(
        features=features,
        labels=labels,
    )
    tr = acquisitions.TrustRegion.build(
        _build_mock_continuous_array_specs(4),
        data=data,
    )
    np.testing.assert_allclose(
        tr.min_linf_distance(
            np.array([
                [0.0, 0.2, 0.3, 0.0],
                [0.9, 0.8, 0.9, 0.9],
                [1.0, 1.0, 1.0, 1.0],
            ]),
        ),
        np.array([0.3, 0.2, 0.0]),
    )
    self.assertAlmostEqual(tr.trust_radius, 0.44, places=3)

  def test_trust_region_padded_small(self):
    # Test that padding still retrieves the same distance computations as
    # `test_trust_region_small`.
    data = types.StochasticProcessModelData(
        features=np.array([
            [0.0, 0.0, 0.0, 0.0, np.nan, np.nan],
            [1.0, 1.0, 1.0, 1.0, np.nan, np.nan],
        ]),
        labels=np.array([0.0, 0.0]),
    )
    tr = acquisitions.TrustRegion.build(
        _build_mock_continuous_array_specs(4),
        data=data,
    )
    data_masked = types.StochasticProcessModelData(
        features=np.array([
            [0.0, 0.0, 0.0, 0.0, np.nan, np.nan],
            [1.0, 1.0, 1.0, 1.0, np.nan, np.nan],
        ]),
        labels=np.array([0.0, 0.0]),
        dimension_is_missing=np.array([False, False, False, False, True, True]),
    )
    tr = acquisitions.TrustRegion.build(
        _build_mock_continuous_array_specs(4),
        data=data_masked,
    )

    np.testing.assert_allclose(
        tr.min_linf_distance(
            np.array([
                [0.0, 0.2, 0.3, 0.0, -100.0, 20.0],
                [0.9, 0.8, 0.9, 0.9, 23.0, 27.0],
                [1.0, 1.0, 1.0, 1.0, 2.0, -3.0],
            ]),
        ),
        np.array([0.3, 0.2, 0.0]),
    )
    self.assertAlmostEqual(tr.trust_radius, 0.224, places=3)


class TrustRegionWithCategoricalTest(absltest.TestCase):

  def test_trust_region_with_categorical(self):
    n_trusted = 20
    n_samples = 5
    d_cont = 6
    d_cat = 3

    trusted = types.ContinuousAndCategoricalArray(
        np.random.randn(n_trusted, d_cont), np.random.randn(n_trusted, d_cat)
    )
    data = types.StochasticProcessModelData(
        features=trusted, labels=np.random.randn(n_trusted)
    )
    xs = types.ContinuousAndCategoricalArray(
        continuous=np.random.randn(n_samples, d_cont),
        categorical=np.random.randn(n_samples, d_cat),
    )
    tr = acquisitions.TrustRegion.build(specs=None, data=data)
    self.assertEqual(tr.min_linf_distance(xs).shape, (n_samples,))

  def test_trust_region_small(self):
    trusted = types.ContinuousAndCategoricalArray(
        continuous=np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ]),
        categorical=np.random.randint(0, 10, size=(2, 4)),
    )
    data = types.StochasticProcessModelData(
        features=trusted, labels=np.random.randn(2)
    )
    tr = acquisitions.TrustRegion.build(specs=None, data=data)
    np.testing.assert_allclose(
        tr.min_linf_distance(
            types.ContinuousAndCategoricalArray(
                continuous=np.array([
                    [0.0, 0.3],
                    [0.9, 0.8],
                    [1.0, 1.0],
                ]),
                categorical=np.random.randint(0, 10, size=(3, 4)),
            )
        ),
        np.array([0.3, 0.2, 0.0]),
    )


if __name__ == '__main__':
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Eagle Strategy Optimizer.

Implements a variation of Eagle Strategy without the Levy random walk, aka
Firefly Algorithm (FA) for the purpose of optimizing a given objective function.

Reference: Yang XS. (2009) Firefly Algorithms for Multimodal Optimization.
In: Stochastic Algorithms: Foundations and Applications (SAGA) 2009.
DOI: https://doi.org/10.1007/978-3-642-04944-6_14

Firefly Algorithm Summary
=========================
FA is a genetic algorithm that maintains a pool of fireflies. Each
firefly emits a light whose intensity is non-decreasing in (or simply equal
to) the objective value. Each iteration, a firefly chases after a brighter
firefly, but the brightness it perceives decreases in distance. This allows
multiple "clusters" to form, as opposed to all fireflies collapsing to a
single point. Not included in the original algorithm, we added "repulsion" which
in addition to the "attraction" forces, meaning fireflies move towards the
bright spots as well as away from the dark spots. We also support non-decimal
parameter types (categorical, discrete, integer), and treat them uniquely when
computing distance, adding pertrubation, and mutating fireflies.

For more details, see the linked paper.

OSS Vizier Implementation Summary
=================================
The fireflies are stored in three JAX arrays: features, rewards, perturbations.
Each iteration we mutate 'batch_size' fireflies to generate new features. The
new features are evaluated on the objective function to obtain their associated
rewards and update the pool where improvement was obtained, and decrease the
perturbation factor otherwise.

If the firefly's perturbation reaches the `perturbation_lower_bound` threshold
it's removed and replaced with new random features.

For performance consideration, the 'pool size' is a multiplier of the
'batch size', and so each iteration the pool is sliced to obtain the current
fireflies to be mutated.

Passing prior trials is supported for knowledge-transfering from previous runs
of the optimizer. Prior trials are used to populate the pool with fireflies
which are closer to the optimum.

Note that the strategy assumes that the search space is not conditional.

Example
=======
# Construct the optimizer with eagle strategy.
optimizer = VectorizedOptimizerFactory(
    VectorizedEagleStrategyFactory(),
)
# Run the optimization.
trials = optimizer.optimize(problem_statement, objective_function)
"""
# pylint: disable=g-long-lambda

import enum
import logging
import math
from typing import Optional, Tuple, Union

import attr
from flax import struct
import jax
from jax import numpy as jnp
from vizier._src.algorithms.optimizers import eagle_param_handler
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import types
from vizier.pyvizier import converters


@enum.unique
class MutateNormalizationType(enum.IntEnum):
  """The force normalization mode. Use IntEnum for JIT compatibility."""

  MEAN = 0
  RANDOM = 1
  UNNORMALIZED = 2


@struct.dataclass
class EagleStrategyConfig:
  """Eagle Strategy optimizer config.

  Attributes:
    visibility: The sensetivity to distance between flies when computing pulls.
    gravity: The maximum amount of attraction pull.
    negative_gravity: The maximum amount of repulsion pull.
    perturbation: The default amount of noise for perturbation.
    categorical_perturbation_factor: A factor to apply on categorical params.
    pure_categorical_perturbation_factor: A factor on purely categorical space.
    perturbation_lower_bound: The threshold below flies are removed from pool.
    penalize_factor: The perturbation decrease for unsuccessful flies.
    pool_size_exponent: An exponent for computing pool size based on search
      space size.
    pool_size: An optional way to set the pool size. If not 0, this takes
      precedent over the programatic pool size computation.
    max_pool_size: Ceiling on pool size.
    mutate_normalization_type: The type of force mutation normalizatoin used for
      damping down the force intensity to stay within reasonable bounds.
    normalization_scale: A mutation force scale-factor to control intensity.
    prior_trials_pool_pct: The percentage of the pool populated by prior trials.
  """

  # Visibility
  visibility: float = 0.45
  # Gravity
  gravity: float = 1.5
  negative_gravity: float = 0.008
  # Perturbation
  perturbation: float = 0.16
  categorical_perturbation_factor: float = 25
  pure_categorical_perturbation_factor: float = 30
  # Penalty
  perturbation_lower_bound: float = 7e-5
  penalize_factor: float = 7e-1
  # Pool size
  pool_size_exponent: float = struct.field(pytree_node=False, default=1.2)
  pool_size: int = struct.field(pytree_node=False, default=0)
  max_pool_size: int = struct.field(pytree_node=False, default=100)
  # Force normalization mode
  mutate_normalization_type: MutateNormalizationType = struct.field(
      pytree_node=False,
      default=MutateNormalizationType.MEAN,
  )
  # Multiplier factor when using normalized modes
  normalization_scale: float = 0.5
  # The percentage of the firefly pool to be populated with prior trials
  prior_trials_pool_pct: float = struct.field(pytree_node=False, default=0.96)


@attr.define(frozen=True)
class VectorizedEagleStrategyFactory(vb.VectorizedStrategyFactory):
  """Eagle strategy factory."""

  eagle_config: EagleStrategyConfig = attr.field(factory=EagleStrategyConfig)

  def __call__(
      self,
      converter: Union[
          converters.TrialToArrayConverter,
          converters.PaddedTrialToArrayConverter,
      ],
      suggestion_batch_size: Optional[int] = None,
  ) -> "VectorizedEagleStrategy":
    """Create a new vectorized eagle strategy.

    In order to create the strategy a converter has to be passed, which the
    strategy will then store a pointer to. The strategy uses the converter to
    get information about the original Vizier Parameters and their relation to
    the array indices (which index belong to which Vizier Parameter). This is
    useful for example for sampling CATEGORICAL features.

    Arguments:
      converter: TrialToArrayConverter that matches the converter used in
        computing the objective / acuisition function.
      suggestion_batch_size: The batch_size of the returned suggestion array.

    Returns:
      A new instance of VectorizedEagleStrategy.
    """
    param_handler = eagle_param_handler.EagleParamHandler.build(
        converter,
        self.eagle_config.categorical_perturbation_factor,
        self.eagle_config.pure_categorical_perturbation_factor,
    )
    n_feature_dimensions_with_padding = converter.to_features([]).shape[-1]
    n_features = len(converter.output_specs)

    if self.eagle_config.pool_size > 0:
      # This allow to override the pool size computation.
      pool_size = self.eagle_config.pool_size
    else:
      pool_size = 10 + int(
          0.5 * n_features + n_features**self.eagle_config.pool_size_exponent
      )
      pool_size = min(pool_size, self.eagle_config.max_pool_size)
      if suggestion_batch_size is not None:
        # If the batch_size was set, ensure pool_size is multiply of batch_size.
        pool_size = int(
            math.ceil(pool_size / suggestion_batch_size) * suggestion_batch_size
        )
    logging.info("Pool size: %d", pool_size)
    if suggestion_batch_size is None:
      # This configuration updates all the fireflies in each iteration.
      suggestion_batch_size = pool_size
    # Use priors to populate Eagle state
    return VectorizedEagleStrategy(
        param_handler=param_handler,
        n_features=n_features,
        n_feature_dimensions=sum(
            spec.num_dimensions for spec in converter.output_specs
        ),
        n_feature_dimensions_with_padding=n_feature_dimensions_with_padding,
        batch_size=suggestion_batch_size,
        config=self.eagle_config,
        pool_size=pool_size,
    )


@struct.dataclass
class VectorizedEagleStrategyState:
  """Container for Eagle strategy state."""

  iterations: jax.Array  # Scalar integer.
  features: jax.Array  # Shape (pool_size, n_features).
  rewards: jax.Array  # Shape (pool_size,).
  best_reward: jax.Array  # Scalar float.
  perturbations: jax.Array  # Shape (pool_size,).


@struct.dataclass
class VectorizedEagleStrategy(vb.VectorizedStrategy):
  """Eagle strategy implementation for maximization problem based on Numpy.

  Attributes:
    converter: The converter used for the optimization problem.
    config: The Eagle strategy configuration.
    n_features: The number of features.
    batch_size: The number of suggestions generated at each suggestion call.
    pool_size: The total number of flies in the pool.
  """

  param_handler: eagle_param_handler.EagleParamHandler
  n_features: int = struct.field(pytree_node=False)
  n_feature_dimensions: int = struct.field(pytree_node=False)
  n_feature_dimensions_with_padding: int = struct.field(pytree_node=False)
  pool_size: int = struct.field(pytree_node=False)
  batch_size: Optional[int] = struct.field(pytree_node=False, default=None)
  config: EagleStrategyConfig = struct.field(
      default_factory=EagleStrategyConfig
  )

  def __post_init__(self):
    logging.info("Eagle class attributes:\n%s", self)
    logging.info("Eagle configuration:\n%s", self.config)

  def init_state(
      self,
      seed: jax.random.KeyArray,
      prior_features: Optional[types.Array] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> VectorizedEagleStrategyState:
    """Initializes the state."""
    if prior_features is not None or prior_rewards is not None:
      init_features = self._populate_pool_with_prior_trials(
          seed, prior_features, prior_rewards
      )
    else:
      init_features = self.param_handler.random_features(self.pool_size, seed)
    return VectorizedEagleStrategyState(
        iterations=jnp.array(0),
        features=init_features,
        rewards=jnp.ones(self.pool_size) * -jnp.inf,
        best_reward=-jnp.inf,
        perturbations=jnp.ones(self.pool_size) * self.config.perturbation,
    )

  def _populate_pool_with_prior_trials(
      self,
      seed: jax.random.KeyArray,
      prior_features: types.Array,
      prior_rewards: types.Array,
  ) -> jax.Array:
    """Populate the pool with prior trials.

    Args:
      seed: Random seed.
      prior_features: (n_prior_features, features_count)
      prior_rewards: (n_prior_features, )

    Returns:
      initial_features

    A portion of the pool is first populated with random features based on
    'prior_trials_pool_pct', then the rest of the flies are populated by
    sequentially iterate over the prior trials, finding the cloest firefly in
    the pool and replace it if the reward is better.
    """
    if prior_features is None or prior_rewards is None:
      raise ValueError("One of prior features / prior rewards wasn't provided!")
    if prior_features.shape[0] != prior_rewards.shape[0]:
      raise ValueError(
          f"prior features shape ({prior_features.shape[0]}) doesn't match"
          f" prior  rewards shape ({prior_rewards.shape[0]})!"
      )
    if prior_features.shape[1] != self.n_feature_dimensions_with_padding:
      raise ValueError(
          f"prior features shape ({prior_features.shape[1]}) doesn't match"
          f" n_features {self.n_feature_dimensions_with_padding}!"
      )

    if len(prior_rewards.shape) > 1:
      raise ValueError("prior rewards is expected to be 1D array!")

    # Reverse the order of prior trials to assign more weight to recent trials.
    flipped_prior_features = jnp.flip(prior_features, axis=0)
    flipped_prior_rewards = jnp.flip(prior_rewards, axis=0)

    # Fill pool with random features.
    n_random_flies = int(
        self.pool_size * (1 - self.config.prior_trials_pool_pct)
    )
    seed1, seed2 = jax.random.split(seed)
    init_features = self.param_handler.random_features(n_random_flies, seed1)
    pool_left_space = self.pool_size - n_random_flies

    if prior_features.shape[0] < pool_left_space:
      # Less prior trials than left space. Take all prior trials for the pool.
      init_features = jnp.concatenate([init_features, flipped_prior_features])
      # Randomize the rest of the pool fireflies.
      random_features = self.param_handler.random_features(
          self.pool_size - len(init_features), seed2
      )
      return jnp.concatenate([init_features, random_features])
    else:
      # More prior trials than left space. Iteratively populate the pool.
      tmp_features = flipped_prior_features[:pool_left_space]
      tmp_rewards = flipped_prior_rewards[:pool_left_space]

      def _loop_body(i, args):
        features, rewards = args
        ind = jnp.argmin(
            jnp.sum(jnp.square(flipped_prior_features[i] - features), axis=-1)
        )
        return jax.lax.cond(
            rewards[ind] < flipped_prior_rewards[i],
            lambda: (
                features.at[ind].set(flipped_prior_features[i]),
                rewards.at[ind].set(flipped_prior_rewards[i]),
            ),
            lambda: (features, rewards),
        )

      # TODO: Use a vectorized method to populate the pool and avoid
      # the for-loop.
      tmp_features, _ = jax.lax.fori_loop(
          lower=pool_left_space,
          upper=prior_features.shape[0],
          body_fun=_loop_body,
          init_val=(tmp_features, tmp_rewards),
      )
      return jnp.concatenate([init_features, tmp_features])

  @property
  def suggestion_batch_size(self) -> int:
    """The number of suggestions returned at each call of 'suggest'."""
    return self.batch_size

  def suggest(
      self, state: VectorizedEagleStrategyState, seed: jax.random.KeyArray
  ) -> jax.Array:
    """Suggest new mutated and perturbed features.

    After initializing, at each call `batch_size` fireflies are mutated to
    generate new features using pulls (attraction/repulsion) from all other
    fireflies in the pool.

    Args:
      state: Current strategy state.
      seed: Random seed.

    Returns:
      suggested batch features: (batch_size, n_features)
    """
    batch_id = state.iterations % (self.pool_size // self.batch_size)
    start = batch_id * self.batch_size
    features_batch = jax.lax.dynamic_slice_in_dim(
        state.features, start, self.batch_size
    )
    rewards_batch = jax.lax.dynamic_slice_in_dim(
        state.rewards, start, self.batch_size
    )
    perturbations_batch = jax.lax.dynamic_slice_in_dim(
        state.perturbations, start, self.batch_size
    )
    features_seed, perturbations_seed, cat_seed = jax.random.split(seed, num=3)

    def _mutate_features(features_batch_):
      mutated_features = self._create_features(
          state.features,
          state.rewards,
          features_batch_,
          rewards_batch,
          features_seed,
      )
      perturbations = self._create_random_perturbations(
          perturbations_batch, perturbations_seed
      )
      return mutated_features + perturbations

    # If the strategy is still initializing, return the random/prior features.
    new_features = jax.lax.cond(
        state.iterations < self.pool_size // self.batch_size,
        lambda x: x,
        _mutate_features,
        features_batch,
    )

    new_features = self.param_handler.sample_categorical(new_features, cat_seed)
    # TODO: The range of features is not always [0, 1].
    #   Specifically, for features that are single-point, it can be [0, 0]; we
    #   also want this code to be aware of the feature's bounds to enable
    #   contextual bandit operation.  Note that if a parameter's bound changes,
    #   we might also want to change the firefly noise or normalizations.
    return jnp.clip(new_features, 0.0, 1.0)

  def _create_features(
      self,
      features: jax.Array,
      rewards: jax.Array,
      features_batch: jax.Array,
      rewards_batch: jax.Array,
      seed: jax.random.KeyArray,
  ) -> jax.Array:
    """Create new batch of mutated and perturbed features.

    The pool fireflies forces (pull/push) are being normalized to ensure the
    combined force doesn't throw the firefly too far. Mathematically, the
    normalization guarantees that the combined normalized force is within the
    simplex constructed by the unnormalized forces and therefore within bounds.

    Args:
      features: (pool_size, n_features)
      rewards: (pool_size,)
      features_batch: (batch_size, n_features)
      rewards_batch: (batch_size,)
      seed: Random seed.

    Returns:
      batch features: (batch_size, n_features)
    """
    # Compute the pairwise squared distances between the features batch and the
    # pool. We use a less numerically precise squared distance formulation to
    # avoid materializing a possibly large intermediate of shape
    # (batch_size, pool_size, n_features).
    dists = (
        jnp.sum(features_batch**2, axis=-1, keepdims=True)
        + jnp.sum(features**2, axis=-1)
        - 2.0 * jnp.matmul(features_batch, features.T)
    )  # shape (batch_size, pool_size)

    # Compute the scaled direction for applying pull between two flies.
    # scaled_directions[i,j] := direction of force applied by fly 'j' on fly
    # 'i'. Note that to compute 'directions' we might perform subtract with
    # removed flies with having value of -np.inf. Moreover, we might even
    # subtract between two removed flies which will result in np.nan. Both cases
    # are handled when computing the actual feautre changes applying a relevant
    # mask.
    directions = rewards - rewards_batch[:, jnp.newaxis]
    scaled_directions = jnp.where(
        directions >= 0.0, self.config.gravity, -self.config.negative_gravity
    )  # shape (batch_size, pool_size)

    # Normalize the distance by the number of features.
    # Get the number of non-padded features.
    force = jnp.exp(
        -self.config.visibility * dists / self.n_feature_dimensions * 10.0
    )
    scaled_force = scaled_directions * force
    # Handle removed fireflies without updated rewards.
    finite_ind = jnp.isfinite(rewards).astype(scaled_force.dtype)

    # Ignore fireflies that were removed from the pool.
    scaled_force = scaled_force * finite_ind

    # Separate forces to pull and push so to normalize them separately.
    scaled_pulls = jnp.maximum(scaled_force, 0.0)
    scaled_push = jnp.minimum(scaled_force, 0.0)

    if self.config.mutate_normalization_type == MutateNormalizationType.MEAN:
      # Divide the push and pull forces by the number of flies participating.
      # Also multiply by normalization_scale.
      norm_scaled_pulls = self.config.normalization_scale * jnp.nan_to_num(
          scaled_pulls / jnp.sum(scaled_pulls > 0.0, axis=1, keepdims=True), 0
      )
      norm_scaled_push = self.config.normalization_scale * jnp.nan_to_num(
          scaled_push / jnp.sum(scaled_push < 0.0, axis=1, keepdims=True), 0
      )
    elif self.config.mutate_normalization_type == (
        MutateNormalizationType.RANDOM
    ):
      # Create random matrices and normalize each row, s.t. the sum is 1.
      pull_seed, push_seed = jax.random.split(seed)
      scaled_pulls_pos = scaled_pulls > 0
      pull_rand_matrix = (
          jax.random.uniform(pull_seed, shape=scaled_pulls.shape)
          * scaled_pulls_pos
      )
      pull_weight_matrix = pull_rand_matrix / jnp.sum(
          pull_rand_matrix, axis=1, keepdims=True
      )
      push_rand_matrix = (
          jax.random.uniform(push_seed, shape=scaled_pulls.shape)
          * scaled_pulls_pos
      )
      push_weight_matrix = push_rand_matrix / jnp.sum(
          push_rand_matrix, axis=1, keepdims=True
      )
      # Normalize pulls/pulls by the weight matrices.
      # Also multiply by normalization_scale.
      norm_scaled_pulls = (
          self.config.normalization_scale * scaled_pulls * pull_weight_matrix
      )
      norm_scaled_push = (
          self.config.normalization_scale * scaled_push * push_weight_matrix
      )
    elif self.config.mutate_normalization_type == (
        MutateNormalizationType.UNNORMALIZED
    ):
      # Doesn't normalize the forces. Use this option with caution.
      norm_scaled_pulls = scaled_pulls
      norm_scaled_push = scaled_push

    # Sums normalized forces (pull/push) of all fireflies. This is equivalent to
    # features_dist[i, j] := distance between fly 'j' and fly 'i'
    # but avoids materializing the large pairwise distance matrix.
    scale = norm_scaled_pulls + norm_scaled_push
    features_changes = jnp.matmul(scale, features) - features_batch * jnp.sum(
        scale, axis=-1, keepdims=True
    )
    return features_batch + features_changes

  def _create_random_perturbations(
      self,
      perturbations_batch: jax.Array,
      seed: jax.random.KeyArray,
  ) -> jax.Array:
    """Create random perturbations for the newly created batch.

    Args:
      perturbations_batch: (batch_size,)
      seed: Random seed.

    Returns:
      perturbations: (batch_size, n_features)
    """
    # Generate normalized noise for each batch.
    batch_noise = jax.random.laplace(
        seed, shape=(self.batch_size, self.n_feature_dimensions_with_padding)
    )
    batch_noise /= jnp.max(jnp.abs(batch_noise), axis=1, keepdims=True)
    return (
        batch_noise
        * perturbations_batch[:, jnp.newaxis]
        * self.param_handler.perturbation_factors
    )

  def update(
      self,
      state: VectorizedEagleStrategyState,
      batch_features: types.Array,
      batch_rewards: types.Array,
      seed: jax.random.KeyArray,
  ) -> VectorizedEagleStrategyState:
    """Update the firefly pool based on the new batch of results.

    Arguments:
      state: Current state.
      batch_features: (batch_size, n_features)
      batch_rewards: (batch_size, )
      seed: Random seed.

    Returns:
      new_state: Updated state.
    """
    new_best_reward = jnp.maximum(state.best_reward, jnp.max(batch_rewards))
    batch_id = state.iterations % (self.pool_size // self.batch_size)
    batch_start_ind = batch_id * self.batch_size
    batch_perturbations = jax.lax.dynamic_slice_in_dim(
        state.perturbations, batch_start_ind, self.batch_size
    )

    def _update(batch_features, batch_rewards, batch_perturbations):
      # Pass the new batch rewards and the associated last suggested features.
      new_batch_features, new_batch_rewards, new_batch_perturbations = (
          self._update_pool_features_and_rewards(
              batch_features,
              batch_rewards,
              jax.lax.dynamic_slice_in_dim(
                  state.features, batch_start_ind, self.batch_size
              ),
              jax.lax.dynamic_slice_in_dim(
                  state.rewards, batch_start_ind, self.batch_size
              ),
              batch_perturbations,
          )
      )
      return self._trim_pool(
          new_batch_features,
          new_batch_rewards,
          new_batch_perturbations,
          new_best_reward,
          seed,
      )

    # If the strategy is still initializing, return the random/prior values.
    (new_batch_features, new_batch_rewards, new_batch_perturbations) = (
        jax.lax.cond(
            state.iterations < self.pool_size // self.batch_size,
            lambda *args: args,
            _update,
            batch_features,
            batch_rewards,
            batch_perturbations,
        )
    )

    return VectorizedEagleStrategyState(
        iterations=state.iterations + 1,
        features=jax.lax.dynamic_update_slice_in_dim(
            state.features, new_batch_features, batch_start_ind, axis=0
        ),
        rewards=jax.lax.dynamic_update_slice_in_dim(
            state.rewards, new_batch_rewards, batch_start_ind, axis=0
        ),
        best_reward=new_best_reward,
        perturbations=jax.lax.dynamic_update_slice_in_dim(
            state.perturbations,
            new_batch_perturbations,
            batch_start_ind,
            axis=0,
        ),
    )

  def _update_pool_features_and_rewards(
      self,
      batch_features: jax.Array,
      batch_rewards: jax.Array,
      prev_batch_features: jax.Array,
      prev_batch_rewards: jax.Array,
      perturbations: jax.Array,
  ) -> Tuple[jax.Array, jax.Array, jax.Array]:
    """Update the features and rewards for flies with improved rewards.

    Arguments:
      batch_features: (batch_size, n_features), new proposed features batch.
      batch_rewards: (batch_size,), rewards for new proposed features batch.
      prev_batch_features: (batch_size, n_features), previous features batch.
      prev_batch_rewards: (batch_size,), rewards for previous features batch.
      perturbations: (batch_size,)

    Returns:
      sliced features, sliced rewards, sliced perturbations
    """
    # Find indices of flies that their generated features made an improvement.
    improve_indx = batch_rewards > prev_batch_rewards
    # Update successful flies' with the associated last features and rewards.
    new_batch_features = jnp.where(
        improve_indx[..., jnp.newaxis], batch_features, prev_batch_features
    )
    new_batch_rewards = jnp.where(
        improve_indx, batch_rewards, prev_batch_rewards
    )
    # Penalize unsuccessful flies.
    new_batch_perturbations = jnp.where(
        improve_indx, perturbations, perturbations * self.config.penalize_factor
    )
    return new_batch_features, new_batch_rewards, new_batch_perturbations

  def _trim_pool(
      self,
      batch_features: jax.Array,
      batch_rewards: jax.Array,
      batch_perturbations: jax.Array,
      best_reward: jax.Array,
      seed: jax.random.KeyArray,
  ) -> Tuple[jax.Array, jax.Array, jax.Array]:
    """Trim the pool by replacing unsuccessful fireflies with new random ones.

    A firefly is considered unsuccessful if its current perturbation is below
    'perturbation_lower_bound' and it's not the best fly seen thus far.
    Random features are created to replace the existing ones, and rewards
    are set to -np.inf to indicate that we don't have values for those feaures
    yet and we shouldn't use them during suggest.

    Args:
      batch_features: (batch_size, n_features)
      batch_rewards: (batch_size,)
      batch_perturbations: (batch_size,)
      best_reward: Best reward seen so far.
      seed: Random seed.

    Returns:
      updated feature, reward, and perturbation batches.
    """
    indx = batch_perturbations < self.config.perturbation_lower_bound
    # Ensure the best firefly is never removed. For optimization purposes,
    # this logic is inside the if statement to be peformed only if needed.
    indx = indx & (batch_rewards != best_reward)

    # Replace fireflies with random features and evaluate rewards.
    random_features = self.param_handler.random_features(self.batch_size, seed)
    new_batch_features = jnp.where(
        indx[..., jnp.newaxis], random_features, batch_features
    )
    new_batch_perturbations = jnp.where(
        indx, self.config.perturbation, batch_perturbations
    )
    # Setting rewards to -inf to filter out those fireflies during suggest.
    new_batch_rewards = jnp.where(indx, -jnp.inf, batch_rewards)
    return new_batch_features, new_batch_rewards, new_batch_perturbations



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.analyzers import state_analyzer
from vizier._src.benchmarks.experimenters import experimenter_factory
from vizier._src.benchmarks.runners import benchmark_runner
from vizier._src.benchmarks.runners import benchmark_state

from absl.testing import absltest


class StateAnalyzerTest(absltest.TestCase):

  def test_curve_conversion(self):
    dim = 10
    experimenter = experimenter_factory.BBOBExperimenterFactory('Sphere', dim)()

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    benchmark_state_factory = benchmark_state.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=experimenter
    )
    num_trials = 20
    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[benchmark_runner.GenerateAndEvaluate()],
        num_repeats=num_trials,
    )

    states = []
    num_repeats = 3
    for i in range(num_repeats):
      bench_state = benchmark_state_factory(seed=i)
      runner.run(bench_state)
      states.append(bench_state)

    curve = state_analyzer.BenchmarkStateAnalyzer.to_curve(states)
    self.assertEqual(curve.ys.shape, (num_repeats, num_trials))

  def test_empty_curve_error(self):
    with self.assertRaisesRegex(ValueError, 'Empty'):
      state_analyzer.BenchmarkStateAnalyzer.to_curve([])

  def test_different_curve_error(self):
    exp1 = experimenter_factory.BBOBExperimenterFactory('Sphere', dim=2)()
    exp2 = experimenter_factory.BBOBExperimenterFactory('Sphere', dim=3)()

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    state1_factory = benchmark_state.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=exp1
    )
    state2_factory = benchmark_state.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=exp2
    )

    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[benchmark_runner.GenerateAndEvaluate()],
        num_repeats=10,
    )

    state1 = state1_factory()
    state2 = state2_factory()
    runner.run(state1)
    runner.run(state2)

    with self.assertRaisesRegex(ValueError, 'must have same problem'):
      state_analyzer.BenchmarkStateAnalyzer.to_curve([state1, state2])


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Library for padding inputs and arrays in order to reduce chances of recompilation."""

import enum
import math

from typing import List, Tuple, Union
import attrs
import jax
import numpy as np


# We don't use internal types here since types depends on this file.
Array = Union[np.ndarray, jax.Array]


class PaddingType(enum.Enum):
  NONE = 1
  MULTIPLES_OF_10 = 2
  POWERS_OF_2 = 3


@attrs.define(frozen=True, kw_only=True)
class PaddingSchedule:
  num_trials: PaddingType = attrs.field(init=True)
  num_features: PaddingType = attrs.field(init=True)


@attrs.define(frozen=True, kw_only=True)
class PaddedArray:
  # Array of shape [N1, ... Nk].
  padded_array: Array = attrs.field(init=True)
  # Mask per dimension padded. List of Arrays of shape [N1], [N2], ..., [Nk].
  is_missing: List[Array] = attrs.field(init=True)

  @property
  def shape(self):
    return self.padded_array.shape


def padded_dimensions(
    dims: List[int], padding_types: List[PaddingType]
) -> List[int]:
  """Returns the padded shape according to `padding_types`."""
  new_dims = []
  for dim, padding_type in zip(dims, padding_types):
    if padding_type == PaddingType.NONE:
      new_dims.append(dim)
    elif padding_type == PaddingType.MULTIPLES_OF_10:
      new_dims.append(int(math.ceil(dim / 10.0)) * 10)
    elif padding_type == PaddingType.POWERS_OF_2:
      if dim == 0:
        new_dims.append(0)
      else:
        new_dims.append(int(2 ** (math.ceil(math.log(dim, 2)))))
    else:
      raise ValueError(f'{padding_type} unexpected.')
  return new_dims


def pad_features(
    features: np.ndarray, padding_schedule: PaddingSchedule
) -> PaddedArray:
  """Pads features in to a `PaddedArray`."""
  num_trials, num_features = features.shape[-2], features.shape[-1]
  padding_types = [
      padding_schedule.num_trials,
      padding_schedule.num_features,
  ]
  new_num_trials, new_num_features = padded_dimensions(  # pylint: disable=unbalanced-tuple-unpacking
      [num_trials, num_features], padding_types
  )
  features_is_missing = np.array(
      [False] * num_features + [True] * (new_num_features - num_features)
  )
  trials_is_missing = np.array(
      [False] * num_trials + [True] * (new_num_trials - num_trials)
  )
  new_features = np.pad(
      features,
      ((0, new_num_trials - num_trials), (0, new_num_features - num_features)),
      constant_values=np.nan,
  )
  return PaddedArray(
      padded_array=new_features,
      is_missing=[trials_is_missing, features_is_missing],
  )


def pad_labels(
    labels: np.ndarray, padding_schedule: PaddingSchedule
) -> PaddedArray:
  """Pads labels in to a `PaddedArray`."""
  num_trials = labels.shape[-2]
  (new_num_trials,) = padded_dimensions(  # pylint:disable=unbalanced-tuple-unpacking
      [num_trials], [padding_schedule.num_trials]
  )
  trials_is_missing = np.array(
      [False] * num_trials + [True] * (new_num_trials - num_trials)
  )
  new_labels = np.pad(
      labels, ((0, new_num_trials - num_trials), (0, 0)), constant_values=np.nan
  )
  return PaddedArray(padded_array=new_labels, is_missing=[trials_is_missing])


def pad_features_and_labels(
    features: np.ndarray,
    labels: np.ndarray,
    padding_schedule: PaddingSchedule,
) -> Tuple[PaddedArray, PaddedArray]:
  """Pads `features` and `labels` according to a `padding_schedule`.

  Args:
    features: `np.ndarray` of shape `[N, D]`.
    labels: `np.ndarray` of shape `[N, T]`.
    padding_schedule: `PaddingSchedule` namedtuple.

  Returns:
    A tuple of `PaddedArray`s representing:
    * padded_features
    * padded_labels
  """
  new_features = pad_features(features, padding_schedule)
  new_labels = pad_labels(labels, padding_schedule)
  return new_features, new_labels



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""High-level wrappers for stochastic process hyperparameter optimizers."""

# TODO: Add optimizers that are frequently used in the literature.

import functools
import time
from typing import Generic, Optional, Protocol, TypeVar

from absl import logging
import attr
import chex
import jax
from jax import numpy as jnp
from jax import random
import jaxopt
import optax
from tensorflow_probability.substrates import jax as tfp
import tree
from vizier._src.jax import stochastic_process_model as sp

tfb = tfp.bijectors

_Params = TypeVar('_Params', bound=chex.ArrayTree)
OptState = chex.ArrayTree


class Setup(Protocol, Generic[_Params]):
  """Set up the model parameters given RNG key."""

  def __call__(self, rng: jax.random.KeyArray) -> _Params:
    """Set up the model parameters given RNG key."""
    pass


class LossFunction(Protocol, Generic[_Params]):
  """Evaluates model params and returns (loss, dict of auxiliary metrics)."""

  def __call__(self, params: _Params) -> tuple[jax.Array, chex.ArrayTree]:
    """Evaluates model params and returns (loss, dict of auxiliary metrics)."""
    pass


class Optimizer(Protocol, Generic[_Params]):
  """Optimizes the LossFunction.

  Example:

  ```python
  setup: Setup = lambda rng: jax.random.uniform(rng, minval=-5, maxval=5)

  def loss_fn(xs):  # satisfies `LossFunction` Protocol
    xs = jax.nn.sigmoid(xs)
    return jnp.cos(xs * 5 * 2 * jnp.pi) * (2 - 5 * (xs - 0.5)**2), dict()

  optimize = optimizers.OptaxTrainWithRandomRestarts(
      optax.adam(5e-3), epochs=500, verbose=True, random_restarts=50)
  optimal_params, metrics = optimize(setup, loss_fn, jax.random.PRNGKey(0))
  ```
  """

  def __call__(
      self,
      setup: Setup[_Params],
      loss_fn: LossFunction[_Params],
      rng: jax.random.KeyArray,
      *,
      constraints: Optional[sp.Constraint] = None,
  ) -> tuple[_Params, chex.ArrayTree]:
    """Optimizes a LossFunction expecting Params as input.

    When constraint bijectors are applied, note that the returned parameters are
    in the constrained space (the parameter domain), not the unconstrained space
    over which the optimization takes place.

    If the Optimizer uses lower and upper bounds, then it is responsible for
    converting `None` bounds to `+inf` or `-inf` as necessary.

    Args:
      setup: Generates initial points.
      loss_fn: Evaluates a point.
      rng: JAX PRNGKey.
      constraints: Parameter constraints.

    Returns:
      Tuple containing optimal input in the constrained space and optimization
      metrics.
    """
    pass


@attr.define
class OptaxTrainWithRandomRestarts(Optimizer[_Params]):
  """Wraps an Optax optimizer.

  It's recommended to use this optimizer with a loss function that normalizes by
  the number of observations. The unnormalized loss function for parameters `p`
  is typically of the form

  ```None
  -gp_likelihood(observations | p) + regularization(p)
  ```

  where regularization may be a negative prior log probability. The likelihood
  term is approximately proportional to the number of observations. As the
  number of observations changes over the course of a study, dividing the loss
  by this number helps ensure that loss values are roughly of the same order of
  magnitude, such that a constant learning rate may be used for gradient-based
  optimizers. Vizier library models make this adjustment automatically.

  Attributes:
    optimizer: Optax optimizer such as `optax.adam(1e-2)`.
    epochs: Number of train epochs.
    verbose: If >=1, logs the train progress. If >=2, logs the gradients.
    random_restarts: Must be positive; number of random initializations for the
      optimization.
    best_n: Number of best values to return from the initializations; must be
      less than or equal to `random_restarts`.
  """

  optimizer: optax.GradientTransformation = attr.field()
  epochs: int = attr.field(kw_only=True)
  verbose: int = attr.field(kw_only=True, default=0, converter=int)
  random_restarts: int = attr.field(kw_only=True, default=32)
  best_n: int = attr.field(kw_only=True, default=1)

  def __attrs_post_init__(self):
    if self.random_restarts < self.best_n:
      raise ValueError(
          f'Cannot generate {self.best_n} results from'
          f' {self.random_restarts} restarts'
      )

  def __call__(
      self,
      setup: Setup[_Params],
      loss_fn: LossFunction[_Params],
      rng: jax.random.KeyArray,
      *,
      constraints: Optional[sp.Constraint] = None,
  ) -> tuple[_Params, chex.ArrayTree]:
    if constraints is None or constraints.bijector is None:
      bijector = None
      unconstrained_loss_fn = loss_fn
    else:
      bijector = constraints.bijector
      unconstrained_loss_fn = lambda x: loss_fn(bijector(x))

    grad_fn = jax.value_and_grad(unconstrained_loss_fn, has_aux=True)

    def _setup_all(rng: jax.random.KeyArray) -> tuple[_Params, OptState]:
      """Sets up both model params and optimizer state."""
      params = setup(rng)
      if bijector is not None:
        params = bijector.inverse(params)
      opt_state = self.optimizer.init(params)
      return params, opt_state

    def _train_step(
        params: _Params, opt_state: OptState
    ) -> tuple[_Params, OptState, chex.ArrayTree]:
      """One train step."""
      (loss, metrics), grads = grad_fn(params)
      logging.log_if(logging.INFO, 'gradients: %s', self.verbose >= 2, grads)
      updates, opt_state = self.optimizer.update(grads, opt_state, params)
      params = optax.apply_updates(params, updates)
      metrics['loss'] = loss
      return params, opt_state, metrics

    if self.random_restarts > 1:
      # Random restarts are implemented via jax.vmap.
      # Note that both setup_all and train_step are vmapped.
      rngs = random.split(rng, self.random_restarts)
      params, opt_state = jax.vmap(_setup_all)(rngs)
      train_step = jax.vmap(_train_step)
    else:
      params, opt_state = _setup_all(rng)
      train_step = _train_step

    logging.info('Initialized parameters. %s',
                 jax.tree_map(lambda x: x.shape, params))

    # See https://jax.readthedocs.io/en/latest/faq.html#buffer-donation.
    train_step = jax.jit(train_step, donate_argnums=[0, 1])
    metrics = []
    for epoch in range(self.epochs):
      params, opt_state, step_metrics = train_step(params, opt_state)
      logging.log_if(
          logging.INFO,
          'Epoch %s: metrics: %s',
          self.verbose >= 1,
          epoch,
          step_metrics,
      )
      metrics.append(step_metrics)

    # Convert `metrics` from a list of dicts to a dict of arrays with leftmost
    # dimension corresponding to train steps.
    outer_treedef = jax.tree_util.tree_structure([0] * self.epochs)
    transposed_metrics = jax.tree_util.tree_transpose(
        outer_treedef, jax.tree_util.tree_structure(step_metrics), metrics
    )
    metrics = jax.tree_util.tree_map(
        jnp.array,
        transposed_metrics,
        is_leaf=lambda x: jax.tree_util.tree_structure(x) == outer_treedef,
    )

    final_losses = metrics['loss'][-1]
    logging.info('Final loss: %s', final_losses)
    if self.random_restarts > 1:
      # Extract the best only.
      argsorted = jnp.argsort(final_losses)
      logging.info('Best loss(es): %s', final_losses[argsorted[: self.best_n]])
      params = jax.tree_map(lambda x: x[argsorted[: self.best_n]], params)
      if self.best_n == 1:
        params = jax.tree_map(functools.partial(jnp.squeeze, axis=0), params)
    if bijector is not None:
      params = bijector(params)
    return params, metrics


@attr.define
class JaxoptLbfgsB(Optimizer[_Params]):
  """Jaxopt's L-BFGS-B optimizer.

  Jaxopt calls the Scipy or Jax implementation of L-BFGS-B.
  L-BFGS-B is a version of L-BFGS that incorporates box constraints on
  parameters.
  https://digital.library.unt.edu/ark:/67531/metadc666315/m2/1/high_res_d/204262.pdf

  Attributes:
    num_line_search_steps: Maximum number of line search steps.
    random_restarts: Must be positive; number of random initializations for the
      optimization.
    best_n: Number of best values to return from the initializations; must be
      less than or equal to `random_restarts`.
     use_scipy: Uses the scipy version of L-BFGS-B. If False, uses the pure JAX
       L-BFGS-B in Jaxopt, which runs on accelerators.
     tol: Tolerance for stopping criteria.
     maxiter: Max number of iterations.
    _speed_test: If True, return speed test results.
  """

  num_line_search_steps: int = attr.field(kw_only=True, default=20)
  random_restarts: int = attr.field(kw_only=True, default=4)
  best_n: int = attr.field(kw_only=True, default=1)
  use_scipy: bool = attr.field(kw_only=True, default=True)
  tol: float = attr.field(kw_only=True, default=1e-8)
  maxiter: int = attr.field(kw_only=True, default=50)
  _speed_test: bool = attr.field(kw_only=True, default=False)

  def __attrs_post_init__(self):
    if self.random_restarts < self.best_n:
      raise ValueError(
          f'Cannot generate {self.best_n} results from'
          f' {self.random_restarts} restarts'
      )

  def __call__(
      self,
      setup: Setup[_Params],
      loss_fn: LossFunction[_Params],
      rng: jax.random.KeyArray,
      *,
      constraints: Optional[sp.Constraint] = None,
  ) -> tuple[_Params, chex.ArrayTree]:
    # L-BFGS-B may be used on unconstrained problems (in which case it is
    # slightly different from L-BFGS, in that it uses the Cauchy point/subspace
    # minimization to choose the line search direction). Bounds must be None or
    # a tuple of size 2. The tuple must contain lower/upper bounds, which may be
    # None or a pytree with the same structure as the model parameters returned
    # by setup (otherwise Jaxopt will raise an error).
    p = setup(jax.random.PRNGKey(0))
    bounds = None if constraints is None else constraints.bounds
    is_leaf = lambda x: jax.tree_util.treedef_is_leaf(  # pylint: disable=g-long-lambda
        jax.tree_util.tree_structure(x)
    )

    def _none_to_inf(b, inf):
      """Converts None bounds to inf or -inf to pass to the optimizer."""
      if b is None:
        b = inf
      if is_leaf(b):
        # Broadcast scalars to the parameter structure.
        return tree.map_structure(lambda x: b * jnp.ones_like(x), p)
      else:
        # For structured bounds, replace `None`s in the structure with `inf`.
        return tree.map_structure(
            lambda b_, x: jnp.ones_like(x) * (inf if b_ is None else b_), b, p
        )

    if bounds is not None:
      lb = _none_to_inf(bounds[0], -jnp.inf)
      ub = _none_to_inf(bounds[1], jnp.inf)
      bounds = (lb, ub)

    loss = lambda p: loss_fn(p)[0]

    if self.use_scipy:
      lbfgsb = jaxopt.ScipyBoundedMinimize(
          fun=loss,
          method='L-BFGS-B',
          maxiter=self.maxiter,
          options={
              'gtol': self.tol,
              'maxls': self.num_line_search_steps,
          },
      )
    else:
      lbfgsb = jaxopt.LBFGSB(
          fun=loss,
          maxls=self.num_line_search_steps,
          tol=self.tol,
          maxiter=self.maxiter,
      )

    metrics = {}
    if self._speed_test:
      start_time = time.time()
      _ = lbfgsb.run(init_params=p, bounds=bounds)
      metrics['compile+runtime'] = time.time() - start_time

    # TODO: Make sure the loss fn isn't retraced for new instances
    # of the optimizer.
    train_times = []
    losses = []
    params = []
    if self.use_scipy:
      logging.info('Using SCIPY L-BFGS-B w/ %d restarts.', self.random_restarts)
      for rng in jax.random.split(rng, num=self.random_restarts):
        p = setup(rng)
        start_time = time.time()

        position, opt_state = lbfgsb.run(init_params=p, bounds=bounds)
        train_times.append(time.time() - start_time)
        losses.append(opt_state.fun_val)
        params.append(position)
        logging.info(
            'Loss: %s, last step time: %s', opt_state.fun_val, train_times[-1]
        )

      all_params = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *params)
    else:
      logging.info('Using JAX L-BFGS-B w/ %d restarts.', self.random_restarts)
      rngs = jax.random.split(rng, num=self.random_restarts)
      init_params = jax.vmap(setup)(rngs)

      start_time = time.time()
      all_params, opt_states = jax.lax.map(
          lambda p: lbfgsb.run(init_params=p, bounds=bounds), init_params
      )
      train_times = time.time() - start_time
      losses = opt_states.value

    losses = jnp.array(losses)
    argsorted = jnp.argsort(losses)
    logging.info('Best loss(es): %s', losses[argsorted[: self.best_n]])
    optimal_params = jax.tree_util.tree_map(
        lambda p: p[argsorted[: self.best_n]], all_params
    )
    if self.best_n == 1:
      optimal_params = jax.tree_map(
          functools.partial(jnp.squeeze, axis=0), optimal_params
      )

    metrics['loss'] = losses
    if self._speed_test:
      metrics['train_time'] = train_times
    return optimal_params, metrics



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Cross-Platform Backend."""
import enum
import getpass
import time

from typing import Any, Dict, Optional, Sequence, Type, Union

from absl import logging
import attrs
import pyglove as pg
from vizier import pyvizier as vz
from vizier._src.pyglove import algorithms
from vizier._src.pyglove import client
from vizier._src.pyglove import constants
from vizier._src.pyglove import converters
from vizier._src.pyglove import core
from vizier._src.pyglove import pythia as pyglove_pythia
from vizier.client import client_abc


TunerPolicy = pyglove_pythia.TunerPolicy
BuiltinAlgorithm = algorithms.BuiltinAlgorithm

ExpandedStudyName = client.ExpandedStudyName
StudyKey = client.StudyKey


class TunerMode(enum.Enum):
  """Mode for Tuner."""

  # Automatic select primary mode when study is new or study TUNER_ID equals
  # to current tuner ID (for failover) or the primary tuner is not accessible
  # via its BNS. Otherwise select secondary mode.
  AUTO = 0

  # Work as primary tuner, which will host pythia service when algorithm is
  # non-Vizier-builtin. When using Vizier built-in algorithms, all tuners are
  # in secondary mode.
  PRIMARY = 1

  # Work as secondary tuner, which can query or stop a tuner task but not
  # hosting pythia service.
  SECONDARY = 2


# Global policy cache.
_global_policy_cache: Dict[StudyKey, TunerPolicy] = dict()


@attrs.define(auto_attribs=False)
class VizierBackend(pg.tuning.Backend):
  """Vizier backend."""

  # Class-level variables.
  default_owner: str = getpass.getuser()
  default_study_prefix: Optional[str] = None
  tuner_cls: Type[client.VizierTuner] = attrs.field()

  # Instance-level variables.
  _tuner: client.VizierTuner
  _name: str = attrs.field()
  _study: client_abc.StudyInterface = attrs.field()
  _converter: converters.VizierConverter = attrs.field()
  _algorithm: pg.geno.DNAGenerator = attrs.field()
  _suggestion_generator: Any = attrs.field()

  def __init__(
      self,
      name: Optional[str],
      group: Union[None, int, str],
      dna_spec: pg.DNASpec,
      algorithm: pg.DNAGenerator,
      metrics_to_optimize: Sequence[str],
      early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = None,
      num_examples: Optional[int] = None,
      study_owner: Optional[str] = None,
      prior_study_ids: Optional[Sequence[str]] = None,
      add_prior_trials: bool = False,
      is_chief: Optional[bool] = None,
  ):
    self._tuner = self.tuner_cls()
    self._algorithm = algorithm
    study_owner = study_owner or self.default_owner
    prior_study_ids = prior_study_ids or tuple()
    name = self._expand_name(name)

    if is_chief:
      mode = TunerMode.PRIMARY
    elif is_chief is None:
      mode = TunerMode.AUTO
    else:
      mode = TunerMode.SECONDARY

    self._converter = converters.VizierConverter.from_dna_spec(
        dna_spec, metrics_to_optimize
    )

    # Load or create study.
    try:
      self._study = self._tuner.load_study(study_owner, name)
      # Study exists.
      if mode == TunerMode.AUTO:
        if self._tuner.ping_tuner(self._get_chief_tuner_id()):
          mode = TunerMode.SECONDARY
        else:
          # NOTE(daiyip): there could be a race condition that multiple workers
          # elect themselves as the new primary, and all of them regard
          # themselves as the elected. When this happens, multiple workers
          # may be hosting the Pythia service for this study. However, the study
          # will use one address (BNS of the latest elected worker) as the
          # Pythia endpoint. Besides, all states are stored in the study,
          # restart of workers will pick up these states from the study.
          # This is done by replaying existing trials to recompute the states of
          # the search algorithm. Therefore, it does not matter which worker is
          # chosen to serve the study, which should work equally well.
          # On the other hand, it is cheap to serve a PythiaService without
          # incoming queries. Therefore, we do not handle this race conditions
          # with expensive distributed locks.
          mode = TunerMode.PRIMARY
          self._register_self_as_primary()
    except KeyError:
      # Study does not exist.
      if mode == TunerMode.SECONDARY:
        self._study = self._wait_for_study(study_owner, name)
      else:
        mode = TunerMode.PRIMARY  # We will make this a chief
        problem = self._converter.problem_or_dummy
        local_tuner_id = self._tuner.get_tuner_id(self._algorithm)
        problem.metadata.ns(constants.METADATA_NAMESPACE)[
            constants.STUDY_METADATA_KEY_TUNER_ID
        ] = local_tuner_id
        self._study = self._tuner.create_study(
            problem, self._converter, study_owner, name, self._algorithm
        )
        if local_tuner_id == self._get_chief_tuner_id():
          # For multi-thread scenario, `local_tuner_id` will be the same for
          # all the worker threads, therefore there is a chance that multiple
          # worker threads consider themselves as PRIMARY. This does not matter
          # since there is only one Pythia service shared across them.
          mode = TunerMode.PRIMARY  # We will make this a chief
          if add_prior_trials:
            # Trials are added to the study directly upon creation.
            trials: Sequence[vz.Trial] = self._load_prior_trials(
                prior_study_ids
            )
            for trial in trials:
              self._study._add_trial(trial)
        else:
          mode = TunerMode.SECONDARY

    # Set up the generator.
    def _suggestion_generator():
      while (
          num_examples is None or len(list(self._study.trials())) < num_examples
      ):
        trials = self._study.suggest(
            count=1, client_id=self._tuner.get_group_id(group)
        )
        if not trials:
          return
        for trial in trials:
          yield core.Feedback(trial, self._converter)

    self._suggestion_generator = _suggestion_generator()

    if mode != TunerMode.PRIMARY or (
        isinstance(self._algorithm, algorithms.PseudoAlgorithm)
    ):
      # nothing more to do
      return

    # Start pythia service.
    self._tuner.start_pythia_service(_global_policy_cache)

    # Set up the policy.
    self._algorithm.setup(dna_spec)
    prior_trials = tuple()
    if not add_prior_trials:
      # Trials are added to the algorithm only.
      prior_trials: Sequence[vz.Trial] = self._load_prior_trials(
          prior_study_ids
      )
    policy = self._create_policy(early_stopping_policy, prior_trials)
    # Connect the service with policy.
    self._register(study_owner, name, policy)

  def _create_policy(
      self,
      early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy],
      prior_trials: Sequence[vz.Trial],
  ) -> TunerPolicy:
    """Creates a pythia policy.

    Args:
      early_stopping_policy:
      prior_trials:

    Returns:
      Policy.
    """
    if prior_trials:

      def get_trial_history(vizier_trials):
        for trial in vizier_trials:
          tuner_trial = core.VizierTrial(self._converter, trial)
          reward = tuner_trial.get_reward_for_feedback(
              self._converter.metrics_to_optimize
          )
          yield (tuner_trial.dna, reward)

      self._algorithm.recover(get_trial_history(prior_trials))

    return TunerPolicy(
        self._tuner.pythia_supporter(self._study),
        self._converter,
        self._algorithm,
        early_stopping_policy=early_stopping_policy,
    )

  def _get_chief_tuner_id(self) -> str:
    metadata = self._study.materialize_problem_statement().metadata.ns(
        constants.METADATA_NAMESPACE
    )
    try:
      return str(metadata[constants.STUDY_METADATA_KEY_TUNER_ID])
    except KeyError as e:
      raise RuntimeError(
          f'Metadata does not exist in study: {self._study.resource_name}'
      ) from e

  def _register_self_as_primary(self) -> str:
    metadata = vz.Metadata()
    tuner_id = self._tuner.get_tuner_id(self._algorithm)
    metadata.ns(constants.METADATA_NAMESPACE)[
        constants.STUDY_METADATA_KEY_TUNER_ID
    ] = self._tuner.get_tuner_id(self._algorithm)
    self._study.update_metadata(metadata)
    self._tuner.use_pythia_for_study(self._study)
    return tuner_id

  def next(self) -> pg.tuning.Feedback:
    """Gets the next tuning feedback object."""
    trial = next(self._suggestion_generator)  # pytype: disable=wrong-arg-types
    return core.Feedback(self._study.get_trial(trial.id), self._converter)

  def _wait_for_study(
      self, owner: str, name: ExpandedStudyName
  ) -> client_abc.StudyInterface:
    """Wait for the study in a loop."""
    while True:
      try:
        return self._tuner.load_study(owner, name)
      except KeyError:
        logging.info(
            'Study %s (owner=%s) does not exist. Retrying after 10 seconds.',
            name,
            owner,
        )
        time.sleep(10)
      except Exception as e:  # pylint:disable=broad-except
        logging.warn(
            'Could not look up study: %s. Retrying after 60 seconds', e
        )
        time.sleep(60)

  def _load_prior_trials(
      self, prior_study_ids: Optional[Sequence[str]]
  ) -> list[vz.Trial]:
    trials = []
    for prior in prior_study_ids:
      trials.extend(
          self._tuner.load_prior_study(prior)
          .trials(vz.TrialFilter(status=vz.TrialStatus.COMPLETED))
          .get()
      )
    return trials

  def _register(
      self, owner: str, name: ExpandedStudyName, policy: TunerPolicy
  ) -> None:
    """Registers the algorithm for a specific study."""
    study_key = StudyKey(owner, name)

    if study_key in _global_policy_cache:
      existing = _global_policy_cache[study_key]
      if existing.algorithm != policy.algorithm:
        raise ValueError(
            f'Different algorithms are used for the same study {study_key!r}. '
            f'Previous: {existing.algorithm!r}, Current: {policy.algorithm!r}.'
        )
      if existing.early_stopping_policy != policy.early_stopping_policy:
        raise ValueError(
            'Different early stopping policy are used for the same study '
            f'{study_key!r}. Previous: {existing.early_stopping_policy!r}, '
            f'Current: {policy.early_stopping_policy!r}.'
        )

    _global_policy_cache[study_key] = policy

  #
  # Class methods.
  #

  @classmethod
  def use_study_prefix(cls, study_prefix: Optional[str]):
    cls.default_study_prefix = study_prefix or ''

  @classmethod
  def _expand_name(cls, name: Optional[str]) -> ExpandedStudyName:
    """Expand the pyglove study name into the full name in Vizier DB.

    Args:
      name: Name as passed into pyglove.

    Returns:
      Study name to use for Vizier interactions.
    """
    components = []
    if cls.default_study_prefix:
      components.append(cls.default_study_prefix)
    if name:
      components.append(name)
    return ExpandedStudyName('.'.join(components))

  @classmethod
  def _get_study_resource_name(cls, name: str) -> str:
    """Use for testing only."""
    return cls.tuner_cls.load_study(
        owner=cls.default_owner,
        name=ExpandedStudyName(name),
    ).resource_name

  @classmethod
  def poll_result(
      cls, name: str, study_owner: Optional[str] = None
  ) -> pg.tuning.Result:
    """Polls result of a study."""
    return core.Result.from_study(
        cls.tuner_cls.load_study(
            study_owner or cls.default_owner, cls._expand_name(name)
        )
    )
